
import pytest

import collections
import random

from pelita import gamestate_filters as gf
from pelita.game import setup_game, play_turn, prepare_bot_state, split_food
from pelita.layout import parse_layout
from pelita.player import stepping_player
import pelita.utils


def make_gamestate():
    def dummy_team(bot, state):
        return bot.position, state

    # get layout
    layout = """
        ##################
        #. ... .##.     y#
        # # #  .  .### #x#
        # # ##.   .      #
        #      .   .## # #
        #a# ###.  .  # # #
        #b     .##. ... .#
        ################## """

    lt_dict = parse_layout(layout)
    game_state = setup_game([dummy_team, dummy_team], layout_dict=lt_dict, max_rounds=1)

    return game_state


def sub_test_noiser(new_bots, old_bots, turn, should_noise, test_other):
    """ sub test function to check if noiser worked

    Parameters
    ----------
    new_bots: bots after noising
    old_bots: bots before noising
    turn: which turn is it now? 0,1,2,3
    should_noise: should the noiser do something right now, or return same bots?
    test_other: true: then it checks if what was meant to happen to other bots happened
                                    and vice versa

    Returns
    -------
    a boolean
    """

    if test_other:
        if not turn % 2:
            # even case
            if should_noise:
                test_bot1 = not old_bots[1] == new_bots[1]
                test_bot3 = not old_bots[3] == new_bots[3]
                return test_bot1 or test_bot3
            else:
                test_bot1 = old_bots[1] == new_bots[1]
                test_bot3 = old_bots[3] == new_bots[3]
                return test_bot1 and test_bot3
        else:
            if should_noise:
                test_bot0 = not old_bots[0] == new_bots[0]
                test_bot2 = not old_bots[2] == new_bots[2]
                return test_bot0 or test_bot2
            else:
                test_bot0 = old_bots[0] == new_bots[0]
                test_bot2 = old_bots[2] == new_bots[2]
                return test_bot0 and test_bot2
    else:
        # test_own should always mean no change
        if turn % 2:
            test_bot0 = old_bots[0] == new_bots[0]
            test_bot2 = old_bots[2] == new_bots[2]
            return test_bot0 and test_bot2
        else:
            test_bot1 = old_bots[1] == new_bots[1]
            test_bot3 = old_bots[3] == new_bots[3]
            return test_bot1 and test_bot3


@pytest.mark.parametrize("bot_id", range(4))
def test_noiser_no_negative_coordinates(bot_id):

    gamestate = make_gamestate()
    old_bots = gamestate["bots"][:]
    walls = gamestate['walls']
    shape = gamestate['shape']
    bot_position = gamestate['bots'][bot_id]
    enemy_group = 1 - (bot_id // 2)
    enemy_positions = gamestate['bots'][enemy_group::2]
    noised = gf.noiser(walls, shape=shape, bot_position=bot_position, enemy_positions=enemy_positions,
                       noise_radius=5, sight_distance=5, rnd=None)
    new_bots = noised["enemy_positions"]
    test_1 = all(item[0] > 0 for item in new_bots)
    test_2 = all(item[1] > 0 for item in new_bots)

    assert test_1 and test_2


def test_noiser_noising_odd_turn1():

    """ It is the odd team's turn, and the noiser should
    work on even bots """

    test_collect_bot0 = []
    test_collect_bot2 = []
    for ii in range(10):

        # we let it run 10 times because it could return the original position,
        # but in most cases it should return a different pos (due to noise)
        gamestate = make_gamestate()
        gamestate["turn"] = 1
        old_bots = gamestate["bots"]
        new_gamestate = prepare_bot_state(gamestate)
        new_bots = new_gamestate['enemy']['bot_positions']
        test_collect_bot0.append((not old_bots[0] == new_bots[0]))
        test_collect_bot2.append((not old_bots[2] == new_bots[1]))

    assert any(test_collect_bot0) or any(test_collect_bot2)


def test_noiser_noising_odd_turn3():

    """ It is the odd team's turn, and the noiser should
    work on even bots """

    test_collect_bot0 = []
    test_collect_bot2 = []
    for ii in range(10):

        # we let it run 10 times because it could return the original position,
        # but in most cases it should return a different pos (due to noise)
        gamestate = make_gamestate()
        gamestate["turn"] = 3
        old_bots = gamestate["bots"]
        new_gamestate = prepare_bot_state(gamestate)
        new_bots = new_gamestate['enemy']['bot_positions']
        test_collect_bot0.append((not old_bots[0] == new_bots[0]))
        test_collect_bot2.append((not old_bots[2] == new_bots[1]))

    assert any(test_collect_bot0) or any(test_collect_bot2)


def test_noiser_noising_even_turn0():

    """ It is the even team's turn, and the noiser should
    work on odd bots """

    test_collect_bot1 = []
    test_collect_bot3 = []
    for ii in range(10):

        # we let it run 10 times because it could return the original position,
        # but in most cases it should return a different pos (due to noise)
        gamestate = make_gamestate()
        gamestate["turn"] = 0
        old_bots = gamestate["bots"]
        new_gamestate = prepare_bot_state(gamestate)
        new_bots = new_gamestate['enemy']['bot_positions']
        test_collect_bot1.append((not old_bots[1] == new_bots[0]))
        test_collect_bot3.append((not old_bots[3] == new_bots[1]))

    assert any(test_collect_bot1) or any(test_collect_bot3)


def test_noiser_noising_even_turn2():

    """ It is the even team's turn, and the noiser should
    work on odd bots """

    test_collect_bot1 = []
    test_collect_bot3 = []
    for ii in range(10):

        # we let it run 10 times because it could return the original position,
        # but in most cases it should return a different pos (due to noise)
        gamestate = make_gamestate()
        gamestate["turn"] = 2
        old_bots = gamestate["bots"]
        new_gamestate = prepare_bot_state(gamestate)
        new_bots = new_gamestate['enemy']['bot_positions']
        test_collect_bot1.append((not old_bots[1] == new_bots[0]))
        test_collect_bot3.append((not old_bots[3] == new_bots[1]))

    assert any(test_collect_bot1) or any(test_collect_bot3)


def test_noiser_not_noising_own_team_even0():

    """ It is the even team's turn, and the noiser should
    not work on own bots """

    gamestate = make_gamestate()
    gamestate["turn"] = 0
    old_bots = gamestate["bots"]
    new_gamestate = prepare_bot_state(gamestate)
    new_bots = new_gamestate['team']['bot_positions']
    test_bot0 = old_bots[0] == new_bots[0]
    test_bot2 = old_bots[2] == new_bots[1]

    assert test_bot0 or test_bot2


def test_noiser_not_noising_own_team_even2():

    """ It is the even team's turn, and the noiser should
    not work on own bots """

    gamestate = make_gamestate()
    gamestate["turn"] = 2
    old_bots = gamestate["bots"]
    new_gamestate = prepare_bot_state(gamestate)
    new_bots = new_gamestate['team']['bot_positions']
    test_bot0 = old_bots[0] == new_bots[0]
    test_bot2 = old_bots[2] == new_bots[1]

    assert test_bot0 or test_bot2


def test_noiser_not_noising_own_team_odd1():

    """ It is the odd team's turn, and the noiser should
    not work on own bots """

    gamestate = make_gamestate()
    gamestate["turn"] = 1
    old_bots = gamestate["bots"]
    new_gamestate = prepare_bot_state(gamestate)
    new_bots = new_gamestate['team']['bot_positions']
    test_bot1 = old_bots[1] == new_bots[0]
    test_bot3 = old_bots[3] == new_bots[1]

    assert test_bot1 or test_bot3


def test_noiser_not_noising_own_team_odd3():

    """ It is the odd team's turn, and the noiser should
    not work on own bots """

    gamestate = make_gamestate()
    gamestate["turn"] = 3
    old_bots = gamestate["bots"]
    new_gamestate = prepare_bot_state(gamestate)
    new_bots = new_gamestate['team']['bot_positions']
    test_bot1 = old_bots[1] == new_bots[0]
    test_bot3 = old_bots[3] == new_bots[1]

    assert test_bot1 or test_bot3


def test_noiser_not_noising_at_noise_radius0():

    """ It is the any team's turn, and the noiser should
    not do anything, because noise radius is 0 """

    test_collect = []
    for tt in range(4):
        gamestate = make_gamestate()
        old_bots = gamestate["bots"]
        gamestate["turn"] = tt
        gamestate["noise_radius"] = 0
        new_gamestate = prepare_bot_state(gamestate)
        new_team_bots = new_gamestate['team']['bot_positions']
        new_enemy_bots = new_gamestate['enemy']['bot_positions']
        if tt % 2 == 0:
            # team 0
            assert old_bots[0] == new_team_bots[0]
            assert old_bots[1] == new_enemy_bots[0]
            assert old_bots[2] == new_team_bots[1]
            assert old_bots[3] == new_enemy_bots[1]
        else:
            # team 1
            assert old_bots[0] == new_enemy_bots[0]
            assert old_bots[1] == new_team_bots[0]
            assert old_bots[2] == new_enemy_bots[1]
            assert old_bots[3] == new_team_bots[1]


@pytest.mark.parametrize("ii", range(30))
def test_noiser_noising_at_noise_radius_extreme(ii):

    """ It is the any team's turn, and the noiser should
    still noise within confines of maze, despite extreme radius """

    gamestate = make_gamestate()
    gamestate["turn"] = random.randint(0, 3)
    team_id = gamestate["turn"] % 2
    old_bots = gamestate["bots"]
    team_bots = old_bots[team_id::2]
    enemy_bots = old_bots[1 - team_id::2]
    noised = gf.noiser(walls=gamestate["walls"],
                       shape=gamestate["shape"],
                       bot_position=gamestate["bots"][gamestate["turn"]],
                       enemy_positions=enemy_bots,
                       noise_radius=50, sight_distance=5, rnd=None)

    assert all(noised["is_noisy"])

    for noised_pos in noised["enemy_positions"]:
        # check that the noised position is legal
        assert noised_pos not in gamestate["walls"]
        assert 0 <= noised_pos[0] < max(gamestate["walls"])[0]
        assert 0 <= noised_pos[1] < max(gamestate["walls"])[1]


@pytest.mark.parametrize('noise_radius, expected', [
    [0, [(3, 3)]], # original position. not noised
    [1, [(2, 3), (3, 3), (4, 3)]],
    [2, [(1, 3), (2, 3), (3, 1), (3, 3), (4, 3), (5, 3)]],
    [5, [(1, 1), (1, 2), (1, 3), (2, 3), (3, 3),
         (4, 3), (5, 3), (6, 3), (7, 3), (7, 2),
         (6, 1), (5, 1), (4, 1), (3, 1)]]
])
def test_uniform_noise_manhattan(noise_radius, expected, test_layout=None):
    # Test how bot 1(x) observes bot 0(a)
    if not test_layout:
        test_layout = (
        """ ##################
            # #.  .  # .    y#
            # #####    ##### #
            #b a  . #  .  .#x#
            ################## """)
    parsed = parse_layout(test_layout)

    position_bucket = collections.defaultdict(int)
    for i in range(200):
        noised = gf.noiser(walls=parsed['walls'],
                            shape=parsed["shape"],
                            bot_position=parsed['bots'][1],
                            enemy_positions=[parsed['bots'][0]],
                            noise_radius=noise_radius)
        if noise_radius == 0:
            assert noised['is_noisy'] == [False]
        else:
            assert noised['is_noisy'] == [True]

        noised_pos = noised['enemy_positions'][0]
        position_bucket[noised_pos] += 1
    assert 200 == sum(position_bucket.values())
    # Since this is a randomized algorithm we need to be a bit lenient with
    # our tests. We check that each position was selected at least once.
    assert set(position_bucket.keys()) == set(expected)


@pytest.mark.parametrize('noise_radius, test_layout', [
    [0, """
        ##################
        # #      # y     #
        # #####    ##### #
        #b a    #      #x#
        ################## """], # original position. not noised
    [1, """
        ##################
        #   .       y    #
        #  .a.        x  #
        #b  .            #
        ################## """], # noised by one
    [1, """
        ##################
        # b          y   #
        #   .         x  #
        #  .a.           #
        ################## """], # noised by one
    [1, """
        ##################
        #  b          y  #
        #.            x  #
        #a.              #
        ################## """], # noised by one
    [1, """
        ##################
        #   b          .a#
        #  x        y   .#
        #                #
        ################## """], # noised by on
    [1, """
        ##################
        # #  b   #   y   #
        # #####    ##### #
        # .a.   #      #x#
        ################## """], # noised by one
    [1, """
        ##################
        # #   b  #    y  #
        # #####    #####.#
        #  x    #      #a#
        ################## """], # noised by one
    [1, """
        ##################
        # #   b  #   .a. #
        # #####    ##### #
        #  x    #    y # #
        ################## """], # noised by one
])
def test_uniform_noise_manhattan_graphical(noise_radius, test_layout):
    # Test how bot 1(x) observes bot 0(a)
    # the expected locations are where the food is placed
    parsed = parse_layout(test_layout)
    expected = parsed['food'] + [parsed['bots'][0]]
    test_uniform_noise_manhattan(noise_radius, expected, test_layout=test_layout)


@pytest.mark.parametrize('test_layout, is_noisy', [
    ["""
        ##################
        # #..  b #     # #
        #.#####y   ##### #
        #.#....  #     # #
        #.#####.   ##### #
        #..a..#..  #  x  #
        #.#####.   ##### #
        #.#....  #     # #
        #.#####    ##### #
        # #..    #     # #
        ################## """, True],
    ["""
        ##################
        # #    y #   ..# #
        # #####b   #####.#
        # #      # ....#.#
        # #####   .#####.#
        #  x  #  ..#..a..#
        # #####   .#####.#
        # #      # ....#.#
        # #####    #####.#
        # #      #   ..# #
        ################## """, True],
    ["""
        ##################
        # #..   y#     # #
        #.##### b  ##### #
        #.#....  #     # #
        #.#####.   ##### #
        #..a..#.. x#     #
        #.#####.   ##### #
        #.#....  #     # #
        #.#####    ##### #
        # #..    #     # #
        ################## """, True],
    ["""
        ##################
        # #..   y#     # #
        #.##### b  ##### #
        #.#....  #     # #
        #.#####.   ##### #
        #..a..#..x #     #
        #.#####.   ##### #
        #.#....  #     # #
        #.#####    ##### #
        # #..    #     # #
        ################## """, True],
        # when we move too close,
        # the noise disappears
    ["""
        ##################
        # #     y#     # #
        # ##### b  ##### #
        # #      #     # #
        # #####    ##### #
        #  a  # x  #     #
        # #####    ##### #
        # #      #     # #
        # #####    ##### #
        # #      #     # #
        ################## """, False],
])
def test_uniform_noise_manhattan_graphical_distance(test_layout, is_noisy):
    # Test how bot 1 observes bot 0
    # the expected locations are where the food is placed
    parsed = parse_layout(test_layout)
    expected = parsed['food'] + [parsed['bots'][0]]

    position_bucket = collections.defaultdict(int)
    NUM_TESTS = 400
    for i in range(NUM_TESTS):
        noised = gf.noiser(walls=parsed['walls'],
                            shape=parsed['shape'],
                            bot_position=parsed['bots'][1],
                            enemy_positions=[parsed['bots'][0]])
                            # use default values for radius and distance
                            # noise_radius=5, sight_distance=5
        assert noised['is_noisy'] == [is_noisy]

        noised_pos = noised['enemy_positions'][0]
        position_bucket[noised_pos] += 1
    assert NUM_TESTS == sum(position_bucket.values())
    # Since this is a randomized algorithm we need to be a bit lenient with
    # our tests. We check that each position was selected at least once.
    assert set(position_bucket.keys()) == set(expected)


def test_uniform_noise_4_bots_manhattan():
    test_layout = (
    """ ##################
        # #. b.  # .     #
        # #####    #####y#
        #   a  . # .  .#x#
        ################## """)
    parsed = parse_layout(test_layout)

    expected_0 = [(1, 1), (1, 2), (1, 3), (2, 3), (3, 3),
                  (4, 3), (5, 3), (6, 3), (7, 3), (7, 2),
                  (7, 1), (6, 1), (5, 1), (4, 1), (3, 1),
                  (8, 2), (8, 3)]
    position_bucket_0 = collections.defaultdict(int)

    expected_2 = [(1, 1), (1, 2), (2, 3), (3, 3), (4, 3),
                  (5, 3), (6, 3), (7, 3), (8, 2), (8, 1),
                  (7, 1), (6, 1), (5, 1), (4, 1), (3, 1),
                  (9, 2), (8, 3), (7, 2), (10, 1)]
    position_bucket_2 = collections.defaultdict(int)

    for i in range(200):
        noised = gf.noiser(walls=parsed['walls'],
                           shape=parsed['shape'],
                           bot_position=parsed['bots'][1],
                           enemy_positions=parsed['bots'][0::2])

        assert noised['is_noisy'] == [True, True]
        position_bucket_0[noised['enemy_positions'][0]] += 1
        position_bucket_2[noised['enemy_positions'][1]] += 1
    assert 200 == sum(position_bucket_0.values())
    assert 200 == sum(position_bucket_2.values())
    # Since this is a randomized algorithm we need to be a bit lenient with
    # our tests. We check that each position was selected at least once.
    assert set(position_bucket_0.keys()) == set(expected_0)
    assert set(position_bucket_2.keys()) == set(expected_2)


def test_uniform_noise_4_bots_no_noise_manhattan():
    test_layout = (
    """ ##################
        # #.  .  # . b   #
        # #####    #####y#
        #  a  . #  .  .#x#
        ################## """)
    parsed = parse_layout(test_layout)

    expected_0 = [(1, 1), (3, 1), (4, 1), (5, 1), (6, 1),
                  (1, 2), (1, 3), (2, 3), (3, 3), (4, 3), (5, 3),
                  (6, 3), (7, 3), (7, 2) ]
    position_bucket_0 = collections.defaultdict(int)

    expected_2 = [(13, 1)]
    position_bucket_2 = {(13, 1) : 0}

    for i in range(200):
        noised = gf.noiser(walls=parsed['walls'],
                           shape=parsed['shape'],
                           bot_position=parsed['bots'][1],
                           enemy_positions=parsed['bots'][0::2])

        assert noised['is_noisy'] == [True, False]
        position_bucket_0[noised['enemy_positions'][0]] += 1
        position_bucket_2[noised['enemy_positions'][1]] += 1
    assert 200 == sum(position_bucket_0.values())
    assert 200 == sum(position_bucket_2.values())
    # Since this is a randomized algorithm we need to be a bit lenient with
    # our tests. We check that each position was selected at least once.
    assert set(position_bucket_0.keys()) == set(expected_0)
    assert set(position_bucket_2.keys()) == set(expected_2)


def test_noise_manhattan_failure():
    test_layout = (
    """ ##################
        ########## . b   #
        ########## #####y#
        ###a###### .  . x#
        ################## """)
    # we test what bot 1 sees
    # bot 0 should not be noised
    # bot 2 should not be noised
    parsed = parse_layout(test_layout)
    expected = parsed['food'] + [parsed['bots'][0]]

    position_bucket = collections.defaultdict(int)
    # check a few times
    for i in range(5):
        noised = gf.noiser(walls=parsed['walls'],
                            shape=parsed['shape'],
                            bot_position=parsed['bots'][1],
                            enemy_positions=parsed['bots'][0::2])

        assert noised['is_noisy'] == [False, False]
        noised_pos = noised['enemy_positions']
        assert noised_pos == parsed['bots'][0::2]

def test_update_food_ages():
    test_layout = (
    """ ##################
        # #.  .  # . b   #
        # #####    #####y#
        #  a  . #  .  .#x#
        ################## """)
    parsed = parse_layout(test_layout)
    food = split_food(parsed['shape'][0], parsed['food'])
    food_age = [{}, {}]

    parsed.update({
        "food": food,
        "food_age": food_age,
    })

    radius = 1
    expected = [{}, {}]
    # nothing should change for either team, the radius is too small
    assert gf.update_food_age(parsed, 0, radius)['food_age'] == expected
    assert gf.update_food_age(parsed, 1, radius)['food_age'] == expected

    radius = 2
    expected_team0 = [{(3, 1): 1},      # team0
                      {}]               # team1
    expected_team1 = [{},               # team0
                      {(14, 3): 1}]     # team1
    # the two teams should get exactly one pellet updated
    assert gf.update_food_age(parsed, 0, radius)['food_age'] == expected_team0
    assert gf.update_food_age(parsed, 1, radius)['food_age'] == expected_team1

    radius = 4
    expected_team0 = [{(3, 1): 1, (6, 3): 1}, {}] # team0
    expected_team1 = [{}, {(14, 3): 1}] # team1
    assert gf.update_food_age(parsed, 0, radius)['food_age'] == expected_team0
    assert gf.update_food_age(parsed, 1, radius)['food_age'] == expected_team1

@pytest.mark.parametrize('radius, team', [
    [0, 0],
    [0, 1],
    [1, 0],
    [1, 1],
    [2, 0],
    [2, 1]
])
def test_shadow_radius(team, radius):
    test_layout = (
    """ ##################
        #................#
        #............y...#
        #....a..........x#
        #..b.............#
        #................#
        ################## """)
    parsed = parse_layout(test_layout)
    # We want to have food below each bot
    parsed['food'].extend(parsed['bots'])
    food = split_food(parsed['shape'][0], parsed['food'])
    food_age = [{}, {}]

    parsed.update({
        "food": food,
        "food_age": food_age,
    })

    check_food = [[
        {(3, 4), (5, 3)}, # radius 0
        {(4, 4), (2, 4), (4, 3), (5, 4), (3, 3), (6, 3), (3, 5), (5, 2)}, # radius 1
        {(6, 2), (5, 5), (5, 1), (4, 2), (6, 4), (1, 4), (7, 3), (2, 3), (4, 5), (3, 2), (2, 5)}, # radius 2
    ],[
        {(13, 2), (16, 3)}, # radius 0
        {(13, 1), (13, 3), (16, 2), (12, 2), (15, 3), (14, 2), (16, 4)}, # radius 1
        {(12, 1), (13, 4), (14, 1), (15, 4), (12, 3), (11, 2), (16, 5), (14, 3), (15, 2), (16, 1)}, # radius 3
    ]]

    # for each team we sum all foods for all radii below the given radius
    shaded_food = set().union(*check_food[team][0:radius + 1])

    # nothing should change for either team, the radius is too small
    assert set(gf.update_food_age(parsed, team, radius)['food_age'][team].keys()) == shaded_food
    # testing pelita.utils.shaded_food as well here
    assert set(pelita.utils.shaded_food(parsed['bots'][team::2], parsed['food'][team], radius)) == shaded_food


# repeat the test 20 times to exercise the randomness of the relocation algorithm
@pytest.mark.parametrize('dummy', range(20))
def test_relocate_expired_food(dummy):
    test_layout = (
    """ ##################
        # #.  .  # . b   #
        # #####    #####y#
        #  a  . #  .  .#x#
        ################## """)
    team_exp_relocate = ( (3, 1), (14, 3) )
    border = (8, 9)
    for team in (0, 1):
        mx = 1
        parsed = parse_layout(test_layout)
        food = split_food(parsed['shape'][0], parsed['food'])
        food_age = [{pos: mx for pos in team_food} for team_food in food]

        parsed.update({
            "food": food,
            "food_age": food_age,
            "rnd" : random.Random(),
        })

        radius = 2

        parsed.update(gf.update_food_age(parsed, team, radius))
        out = gf.relocate_expired_food(parsed, team, radius, mx)

        # check that the expired pellet is gone, bot only when it's our team turn
        assert team_exp_relocate[team] not in out['food'][team]
        assert team_exp_relocate[1-team] in out['food'][1-team]

        # check that the relocation was done properly
        assert len(parsed['food'][team].intersection(out['food'][team])) == 2
        new = out['food'][team].difference(parsed['food'][team])
        assert len(new) == 1 # there is only one new pellet
        new = new.pop()
        assert new not in parsed['walls'] # it was not located on a wall
        assert new[0] not in border # it was not located on the border
        assert new not in parsed['bots'] # it was not located on a bot
        for team_bot in parsed['bots'][team::2]:
            # it was not located within the shadow of a team bot
            assert gf.manhattan_dist(new, team_bot) > radius
        # check that the new pellet is in the right homezone
        if team == 0:
            assert 0 < new[0]
            assert new[0] < border[0]
        else:
            assert border[1] < new[0]
            assert new[0] < border[0]*2

def test_relocate_expired_food_nospaceleft():
    test_layout = (
    """ ##################
        ###..... # . b   #
        #######y    ######
        ###a##..#  .  .#x#
        ################## """)

    # location of the food pellet that will be relocated
    to_relocate = (3, 1)
    # location of the pellet that will be removed
    to_remove = (7, 1)
    max_age = 2

    parsed = parse_layout(test_layout)
    food = split_food(parsed['shape'][0], parsed['food'])
    food_age = [{}, {}]
    food_age[0][to_relocate] = max_age

    parsed.update({
        "food": food,
        "food_age": food_age,
        "rnd" : random.Random(),
    })

    radius = 2

    parsed.update(gf.update_food_age(parsed, 0, radius))
    out = gf.relocate_expired_food(parsed, 0, radius, max_age)
    # check that the food pellet did not move: there is no space to move it
    # anywhere
    assert to_relocate in out['food'][0]
    assert len(out['food'][0]) == 7
    # check that food age has increased over the allowed maximum
    assert out['food_age'][0] == {to_relocate: max_age + 1}

    # now make space for the food and check that it gets located in the free spot
    parsed.update(out)
    parsed['food'][0].remove(to_remove)
    out = gf.relocate_expired_food(parsed, 0, radius, max_age)
    assert to_relocate not in out['food'][0]
    assert to_remove in out['food'][0]
    # Both the relocated and the moved to position should have their ages reset
    assert to_relocate not in out['food_age'][0]
    assert to_remove not in out['food_age'][0]

def test_pacman_resets_age():
    # We move bot a across the border
    # Once it becomes a bot, the food_age should reset
    test_layout = (
    """ ##################
        #         x     y#
        #       ..       #
        #b     a         #
        ################## """)
    team1 = stepping_player('>>-', '---')
    team2 = stepping_player('<<-', '---')

    parsed = parse_layout(test_layout)
    state = setup_game([team1, team2], layout_dict=parsed, max_rounds=8)
    assert state['food_age'] == [{}, {}]
    state = play_turn(state)
    assert state['turn'] == 0
    assert state['food_age'] == [{(8, 2): 1}, {}]
    state = play_turn(state)
    assert state['turn'] == 1
    assert state['food_age'] == [{(8, 2): 1}, {(9, 2): 1}]
    state = play_turn(state)
    state = play_turn(state)
    state = play_turn(state)
    # NOTE: food_ages are calculated *before* the move
    # Therefore this will only be updated once it is team1’s turn again
    assert state['turn'] == 0
    assert state['food_age'] == [{(8, 2): 3}, {(9, 2): 2}]
    state = play_turn(state)
    assert state['food_age'] == [{(8, 2): 3}, {(9, 2): 3}]
    state = play_turn(state)
    assert state['turn'] == 2
    assert state['food_age'] == [{}, {(9, 2): 3}]
    state = play_turn(state)
    assert state['turn'] == 3
    assert state['food_age'] == [{}, {}]



"""Detect if a layout contains "chambers" with food"""
import sys

import networkx as nx
from pelita.layout import parse_layout
from pelita.utils import walls_to_graph


if __name__ == '__main__':

    # either read a file or read from stdin
    if len(sys.argv) == 1:
        flname = 'stdin'
        layout = parse_layout(sys.stdin.read())
    else:
        flname = sys.argv[1]
        with open(flname, "rt") as fl:
            layout = parse_layout(fl.read())

    # first of all, check for symmetry
    # if something is found on the left, it should be found center-mirrored on the right
    width, height = layout['shape']
    known = set(layout['walls']) | set(layout['food'])
    layout['empty'] = [(x,y) for x in range(width) for y in range(height) if (x,y) not in known]
    for x in range(width // 2):
        for y in range(height):
            coord = (x, y)
            cmirror = (width-x-1, height-y-1)
            for typ_ in ('walls', 'food', 'empty'):
                if (coord in layout[typ_]) and (cmirror not in layout[typ_]):
                        print(f'{flname}: Layout is not symmetric {coord} != {cmirror}')


    graph = walls_to_graph(layout['walls'])

    # check for dead_ends
    for node, degree in graph.degree():
        if degree < 2:
            print(f'{flname}: found dead end in {node}')

    if nx.node_connectivity(graph) < 2:
        entrance = nx.minimum_node_cut(graph).pop()
        print(f'{flname}: Detected chamber, entrance: {entrance}')




#!/usr/bin/env python3

from dataclasses import dataclass
import json
import logging
import random
import queue
import shlex
import signal
import subprocess
import sys
import time
from typing import Optional, Dict, List
import urllib
from urllib.parse import urlparse
import urllib.parse
from weakref import WeakValueDictionary

import click
from rich import print as pprint
from rich.progress import Progress, SpinnerColumn, MofNCompleteColumn, BarColumn, TextColumn, TimeElapsedColumn, Task
import yaml
import zeroconf
import zmq

from ..network import PELITA_PORT
from .script_utils import start_logging

_logger = logging.getLogger(__name__)

zeroconf.log.setLevel(logging.INFO)
zeroconf.log.addHandler(_logger)

DEFAULT_MAX_CONNECTIONS = 50

# TODO: This timeout should be shorter
# once a connection has been established
SEND_QUEUE_TIMEOUT = 5. # seconds

@dataclass
class GameInfo:
    round: Optional[int]
    max_rounds: int
    my_index: int
    my_name: str
    my_score: int
    enemy_name: str
    enemy_score: int
    finished: bool = False
    last_msg: Optional[bytes] = None

    def status(self):
        plural = "" if self.round == 1 else "s"
        finished = f"[b]Finished[/b] ({self.round if self.round is not None else '-'} round{plural}): " if self.finished else ""
        my_name = self.my_name or "[dim]unknown[/dim]"
        enemy_name = self.enemy_name or "[dim]unknown[/dim]"
        if self.my_index == 0:
            return f"{finished}[u]{my_name}[/u] ({self.my_score}) vs {enemy_name} ({self.enemy_score})"
        else:
            return f"{finished}{enemy_name} ({self.enemy_score}) vs [u]{my_name}[/u] ({self.my_score})"

@dataclass(frozen=True)
class ProcessInfo:
    proc: subprocess.Popen
    task: Task
    info: GameInfo
    dealer_id: bytes
    pair_socket: zmq.Socket

@dataclass
class TeamInfo:
    spec: str
    team_name: str
    zeroconf_name: str = None
    server_path: str = None
    team_name_override: str = None
    silent_bots: bool = False


def zeroconf_register(zc, address, port, team_spec, path, print=print):
    parsed_url = urlparse("tcp://" + address + ":" + str(port))
    if parsed_url.scheme != "tcp":
        _logger.warning("Can only advertise to tcp addresses.")
        return
    if parsed_url.hostname == "0.0.0.0":
        _logger.warning("Can only advertise to a specific interface.")
        return

    name = _check_team(team_spec)

    desc = {
        'spec': team_spec,
        'team_name': name,
        'proto_version': 0.2,
        'path': '/' + path,
    }

    # there is a chance that our name is already taken.
    # We try for a few times with different names before we give up
    suffixes = ["", "-1", "-2", "-3", "-4"]

    for suffix in suffixes:
        full_name = f"{name}{suffix}._pelita-player._tcp.local."
        info = zeroconf.ServiceInfo(
            "_pelita-player._tcp.local.",
            full_name,
            parsed_addresses=[parsed_url.hostname],
            #server='mynewserver.local',
            port=parsed_url.port,
            properties=desc,
        )

        print(f"Registration of service {full_name}")
        try:
            zc.register_service(info)
            return info
        except zeroconf.NonUniqueNameException as e:
            print(f"Name {full_name} already taken. Trying alternative")
            continue

    print(f"No alternative found after {len(suffixes)} tries. Not advertising the player.")
    return None

def zeroconf_deregister(zc: zeroconf.Zeroconf, info: zeroconf.ServiceInfo):
    zc.unregister_service(info)

class PelitaServer:
    # TODO: Explain how ROUTER-DEALER works with ZMQ

    def __init__(self, team_infos: List[TeamInfo], address, port, *, advertise: str, session_key: str,
                 max_connections: int):

        self.team_infos = team_infos

        self.address = address
        self.port = port
        self.advertise = advertise
        self.session_key = session_key
        self.max_connections = max_connections

        # maps socket/process/game data
        self.connection_map: Dict[bytes, ProcessInfo] = {}
        self.connections_by_pair_socket = WeakValueDictionary() # automatic cache

        # maps team_spec, path to zc.ServiceInfo
        self.team_serviceinfo_mapping = {}

        def cleanup(_signum, _frame):
            for process_info in self.connection_map.values():
                _logger.warning(f"Terminating unfinished process: ‘{shlex.join(process_info.proc.args)}’.")
                process_info.proc.terminate()
            finish_time = time.monotonic() + 3
            for process_info in self.connection_map.values():
                # We need to wait for all processes to finish
                # Otherwise we might exit before the signal has been sent
                _logger.debug(f"Waiting for process ‘{shlex.join(process_info.proc.args)}’ to terminate.")
                remainder = finish_time - time.monotonic()
                if remainder > 0:
                    try:
                        process_info.proc.wait(remainder)
                    except subprocess.TimeoutExpired:
                        _logger.warning(f"Process ‘{shlex.join(process_info.proc.args)}’ has not finished.")

            sys.exit()

        signal.signal(signal.SIGTERM, cleanup)
        signal.signal(signal.SIGINT, cleanup)

        self.ctx = zmq.Context()
        self.router_sock = self.ctx.socket(zmq.ROUTER)
        self.router_sock.bind(f"tcp://{self.address}:{self.port}")

        self.poll = zmq.Poller()
        self.poll.register(self.router_sock, zmq.POLLIN)

        self.ticks_progressbar = 0.0
        self.ticks_process_cleanup = 0.0

        self.send_queue = queue.SimpleQueue()

    def handle_send_queue(self):
        if self.send_queue.qsize() == 0:
            # The check for the qsize is not reliable and we only
            # use it as an optimisation to skip early
            return

        # Handle all unsent messages to pair sockets
        unsent = set()
        try:
            while True:
                data = self.send_queue.get(block=False)
                time_monotonic, dealer_id, message = data
                if time.monotonic() - time_monotonic > SEND_QUEUE_TIMEOUT:
                    # discard
                    _logger.warning(f"Could not send to dealer id {dealer_id.hex()}.")
                    continue

                if dealer_id not in self.connection_map:
                    _logger.warning(f"Could not send to dealer id {dealer_id.hex()}.")
                    continue

                process_info = self.connection_map[dealer_id]
                process_info.info.last_msg = message

                # Problem: When the receiving end of the pair socket has crashed, then
                # a simple send will halt forever.
                if process_info.pair_socket.poll(0, flags=zmq.POLLOUT) == zmq.POLLOUT:
                    process_info.pair_socket.send(message)
                else:
                    unsent.add(data)

        except queue.Empty:
            for data in unsent:
                self.send_queue.put(data)

    def handle_known_client(self, dealer_id, message, progress):
        # We try to send to the pair socket immediately
        # If this fails, we enqueue the message and send it again later

        process_info = self.connection_map[dealer_id]
        process_info.info.last_msg = message

        try:
            process_info.pair_socket.send(message, flags=zmq.NOBLOCK)
        except zmq.ZMQError:
            data = time.monotonic(), dealer_id, message
            self.send_queue.put(data)
        return


    def handle_new_connection(self, dealer_id, message, progress):
        try:
            msg_obj = json.loads(message)
        except ValueError as e:
            _logger.debug(f"Error {e!r} when parsing incoming message. Ignoring.")
            return

        # TODO actions
        # stop server - do not accept new requests
        # purge server - drop all running connections

        if "STATUS" in msg_obj:
            # check key
            if not msg_obj.get('key', None) == self.session_key:
                return

            progress.console.print("List of teams:")
            for team in self.team_infos:
                progress.console.print(team)

        elif "TEAM" in msg_obj:
            # check key
            if not msg_obj.get('key', None) == self.session_key:
                return

            team_spec = msg_obj.get('team_spec')

            if msg_obj.get('TEAM') == 'ADD':
                team_info = load_team_info(team_spec)
                self.team_infos.append(team_info)
                info = zeroconf_register(self.zc, self.advertise, self.port, team.spec, team.server_path, print=progress.console.print)
                if info:
                    self.team_serviceinfo_mapping[(team.spec, team.server_path)] = info

            if msg_obj.get('TEAM') == 'REMOVE':
                # TODO: cannot remove from self.team_infos yet
                info = self.team_serviceinfo_mapping[(team.spec, team.server_path)]
                zeroconf_deregister(self.zc, info)

        elif "SCAN" in msg_obj:
            # return list of available bots
            if len(self.connection_map) >= self.max_connections:
                _logger.warning("Exceeding maximum number of connections. Ignoring")
                self.router_sock.send_multipart([dealer_id, b"NOCONN"])
                return

            requested_url = urlparse(msg_obj['SCAN'])
            progress.console.log(f"SCAN from id {dealer_id.hex()}: {requested_url.scheme}://{requested_url.hostname}{requested_url.path}")

            avaliable_teams = {}
            for team_info in self.team_infos:
                # we construct the url from the url that reached us
                port = "" if requested_url.port == PELITA_PORT or requested_url.port is None else f":{requested_url.port}"
                full_url = f"{requested_url.scheme}://{requested_url.hostname}{port}/{team_info.server_path}"
                avaliable_teams[full_url] = team_info.team_name

            avaliable_teams_json = json.dumps(avaliable_teams).encode("utf8")

            self.router_sock.send_multipart([dealer_id, avaliable_teams_json])

        elif "REQUEST" in msg_obj:
            # incoming message is a new request
            if len(self.connection_map) >= self.max_connections:
                _logger.warning("Exceeding maximum number of connections. Ignoring")
                self.router_sock.send_multipart([dealer_id, b"NOCONN"])
                return

            # TODO: Do not update status with every message

            requested_url = urlparse(msg_obj['REQUEST'])
            progress.console.log(f"Request from id {dealer_id.hex()}: {requested_url.scheme}://{requested_url.hostname}{requested_url.path}")

            if len(self.team_infos) == 0:
                self.router_sock.send_multipart([dealer_id, b"NOTEAM"])
                return

            # Select default team in case we don’t find any
            team = self.team_infos[0]
            for team_info in self.team_infos:
                if requested_url.path == '/' + team_info.server_path:
                    team = team_info
                    break
            else:
                # not found. use default but warn
                progress.console.print(f"Player for path {requested_url.path} not found. Using default.")

            info = GameInfo(round=None, max_rounds=None, my_index=0,
                            my_name="waiting", my_score=0,
                            enemy_name="waiting", enemy_score=0)
            task = progress.add_task(info.status(), total=info.max_rounds)


            num_running = len(self.connection_map)
            _logger.info(f"Starting match for team {team.spec}. ({num_running} already running.)")
            subproc, pair_sock = run_team_in_subprocess(self.ctx, team.spec, silent_bots=team.silent_bots)

            self.poll.register(pair_sock, zmq.POLLIN)

            process_info = ProcessInfo(proc=subproc, task=task, info=info, dealer_id=dealer_id, pair_socket=pair_sock)
            self.connection_map[process_info.dealer_id] = process_info
            self.connections_by_pair_socket[process_info.pair_socket] = process_info


            # Send a reply to the requester (that the process has started)
            # Otherwise they might already start querying for the team name
            self.router_sock.send_multipart([dealer_id, b"OK"])

        else:
            _logger.info("Unknown incoming DEALER and not a request.")


    def update_progress_bar(self, progress, process_info: ProcessInfo, force_exit=False):
        if not process_info.info.last_msg:
            return

        if process_info.info.finished:
            return

        try:
            msg_obj = json.loads(process_info.info.last_msg)
        except ValueError as e:
            progress.console.log(f"Error {e!r} when parsing incoming message. Ignoring.")
            _logger.warning(f"Error {e!r} when parsing incoming message. Ignoring.")
            return

        try:
            process_info.info.round = msg_obj['__data__']['game_state']['round']
            process_info.info.max_rounds = msg_obj['__data__']['game_state']['max_rounds']
        except KeyError:
            process_info.info.round = None

        if round is not None:
            progress.update(process_info.task, completed=process_info.info.round, total=process_info.info.max_rounds)

        try:
            process_info.info.my_index = msg_obj['__data__']['game_state']['team']['team_index']
            process_info.info.my_name = msg_obj['__data__']['game_state']['team']['name']
            process_info.info.my_score = msg_obj['__data__']['game_state']['team']['score']
            process_info.info.enemy_name = msg_obj['__data__']['game_state']['enemy']['name']
            process_info.info.enemy_score = msg_obj['__data__']['game_state']['enemy']['score']

            progress.update(process_info.task, description=process_info.info.status())
        except KeyError:
            pass

        try:
            is_exit = msg_obj['__action__'] == 'exit'
        except KeyError:
            is_exit = False

        if is_exit or force_exit:
            progress.stop_task(process_info.task)
            progress.update(process_info.task, visible=False)
            process_info.info.finished = True
            progress.console.print(process_info.info.status())


    def start(self):
        zc = zeroconf.Zeroconf()

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
        ) as progress:

            if self.advertise:

                for team in self.team_infos:
                    info = zeroconf_register(zc, self.advertise, self.port, team.spec, team.server_path, print=progress.console.print)
                    if info:
                        self.team_serviceinfo_mapping[(team.spec, team.server_path)] = info

            while True:
                # If we have active connections, we break out every second
                # to update the status bar
                # Otherwise, we’ll just sleep
                poll_timeout = 1000 if len(self.connection_map) else None

                incoming_evts = dict(self.poll.poll(poll_timeout))
                has_router_sock = incoming_evts.pop(self.router_sock, None)

                if has_router_sock == zmq.POLLIN:
                    try:
                        dealer_id, message = self.router_sock.recv_multipart()

                        # check if we know the dealer already
                        if dealer_id in self.connection_map.keys():
                            # incoming message refers to an existing connection
                            self.handle_known_client(dealer_id, message, progress)

                        else:
                            # a new connection. we parse the message and check what we need to do
                            self.handle_new_connection(dealer_id, message, progress)

                    except Exception as e:
                        _logger.debug(f"Error {e!r} when handling incoming message {message}. Ignoring.")

                # Are there any non-router messages waiting for us?
                if len(incoming_evts):
                    # One or more of our spawned players has replied
                    # try to find the according process info
                    for socket in incoming_evts:
                        process_info = self.connections_by_pair_socket.get(socket)
                        if process_info:
                            # success
                            message = process_info.pair_socket.recv()
                            # route message back
                            self.router_sock.send_multipart([process_info.dealer_id, message])

                self.handle_send_queue()

                # not every event needs to update the progress bars
                if (now := time.monotonic()) - self.ticks_progressbar > 0.01:
                    self.ticks_progressbar = now
                    for process_info in list(self.connection_map.values()):
                        self.update_progress_bar(progress, process_info)

                if (now := time.monotonic()) - self.ticks_process_cleanup > 3:
                    self.ticks_process_cleanup = now
                    count = 0
                    for process_info in list(self.connection_map.values()):
                        # check if the process has terminated
                        if process_info.proc.poll() is not None:
                            self.update_progress_bar(progress, process_info, force_exit=True)
                            # We need to unregister the socket or else the polling will take longer and longer
                            self.poll.unregister(process_info.pair_socket)
                            del self.connection_map[process_info.dealer_id]
                            count += 1
                    if count:
                        plural = "" if count == 1 else "es"
                        progress.console.log(f"Cleaned up {count} process{plural}. ({len(self.connection_map)} still running.)")

def load_team_info(team_spec: str) -> Optional[TeamInfo]:
    # Takes a team_spec, tries to run it and returns a team info object

    # TODO: Improve path handling for manual override and duplicate detection

    team_name = _check_team(team_spec)
    if not team_name:
        pprint(f"Team {team_spec} did not return a filename. Skipping.")
        return

    team_info = TeamInfo(team_spec, team_name)

    team_path = team_name.replace(" ", "_")
    team_path = urllib.parse.quote(team_path)

    team_info.server_path = team_path
    pprint(f"Mapping team {team_info.spec} ({team_info.team_name}) to path {team_info.server_path}")
    return team_info

def run_team_in_subprocess(ctx, team_spec, silent_bots=False):
    pair_sock = ctx.socket(zmq.PAIR)
    port = pair_sock.bind_to_random_port('tcp://127.0.0.1')
    pair_addr = 'tcp://127.0.0.1:{}'.format(port)

    subproc = play_remote(team_spec, pair_addr, silent_bots=silent_bots)

    return subproc, pair_sock

# TODO: This could optionally run in a sandbox (systemd-run)
def play_remote(team_spec, pair_addr, silent_bots=False):
    external_call = [sys.executable,
                    '-m',
                    'pelita.scripts.pelita_player',
                    'remote-game',
                    team_spec,
                    pair_addr,
                    *(['--silent-bots'] if silent_bots else []),
                    ]
    _logger.debug("Executing: %r", external_call)
    sub = subprocess.Popen(external_call)
    return sub

# TODO: This could optionally run in a sandbox (systemd-run)
def _check_team(team_spec):
    external_call = [sys.executable,
                    '-m',
                    'pelita.scripts.pelita_player',
                    'check-team',
                    team_spec]
    _logger.debug("Executing: %r", external_call)
    res = subprocess.run(external_call, capture_output=True, text=True)
    return res.stdout.strip()

@click.group()
@click.option('--log',
              is_flag=False, flag_value="-", default=None, metavar='LOGFILE',
              help="print debugging log information to LOGFILE (default 'stderr')")
def main(log):
    if log is not None:
        start_logging(log)

def configure(ctx, param, filename):
    if not filename:
        return
    settings = yaml.load(filename, Loader=yaml.SafeLoader)
    if 'teams' in settings:
        settings['teams'] = [team['spec'] for team in settings['teams'] if team.get('spec')]

    ctx.default_map = settings

@main.command(help="Start a pelita server with given players")
@click.option('--config',
              default=None,
              type=click.File('r'),
              help='Configuration file',
              callback=configure,
              is_eager=True,
              expose_value=False,
              show_default=True)
@click.option('--address', default="0.0.0.0")
@click.option('--port', default=PELITA_PORT)
@click.option('--team', 'teams', type=str, multiple=True, help="Team path")
@click.option('--advertise', default=None, type=str,
              help='advertise player on zeroconf')
@click.option('--max-connections', default=DEFAULT_MAX_CONNECTIONS, show_default=True,
              help='Maximum number of connections that we want to handle')
def remote_server(address, port, teams, advertise, max_connections):
    # When used with --config the following yaml format is expected:
    #
    #    address: 0.0.0.0
    #    port: 41736
    #    teams:
    #    - name: abc
    #      spec: pelita/player/StoppingPlayer
    #    - name: def
    #      spec: path/to/module

    session_key = "".join(str(random.randint(0, 9)) for _ in range(12))
    pprint(f"Use --session-key {session_key} to for the admin API.")

    team_infos = []
    for team_spec in teams:
        team_info = load_team_info(team_spec)
        if team_info:
            team_infos.append(team_info)

    server = PelitaServer(team_infos, address, port, advertise=advertise, session_key=session_key,
                          max_connections=max_connections)
    server.start()

    # asyncio repl …
    # reload via zqm key

def send_api_message(url, session_key, type, subtype, **payload):
    ctx = zmq.Context()
    sock = ctx.socket(zmq.DEALER)
    parsed_url = urlparse(url)

    if parsed_url.scheme not in ['pelita', 'tcp'] :
        raise ValueError('Scheme must be pelita or tcp')

    address = parsed_url.hostname
    port = parsed_url.port or PELITA_PORT

    sock.connect(f"tcp://{address}:{port}")
    sock.send_json({
        type: subtype,
        'key': session_key,
        **payload
    })


@main.command(help="Show server statistics")
@click.option('--url', default="pelita://localhost/")
@click.option('--session-key', type=str, required=True)
def show_statistics(url, session_key):
    send_api_message(url, session_key, "STATUS", "show-stats")

@main.command(help="Add team")
@click.argument('team_spec')
@click.option('--url', default="pelita://localhost/")
@click.option('--session-key', type=str, required=True)
def add_team(url, session_key, team_spec):
    send_api_message(url, session_key, "TEAM", "ADD", team_spec=team_spec)


if __name__ == '__main__':
    main()



from collections import Counter
import itertools
import random
from typing import List, Tuple

def create_matchplan(teams: List[str]) -> List[Tuple[str, str]]:
    """ Takes a list of team ids and returns a list of tuples of team ids with
    the matchplan."""

    # We want to optimise the matchplan in multiple ways:
    # a) At any point each team should have a similar number of played games
    # b) No team should play two games directly after another;
    # both of these points should make the tournament more interesting to
    # follow.
    # Finally, in the interest of fairness:
    # c) Blue and red teams (ie. who starts the match) should be assigned evenly.
    #
    # For an even number of teams (> 4), these points can be fulfilled with the
    # circle method; for an odd number of teams, however, it should be noted
    # that the circle method is not the most optimal algorithm. [1]
    #
    # As having five teams is the usual case during ASPP student tournaments,
    # we will special-case this one and implement the circle method in all
    # other cases. (Arguably, round-robin tournaments with more than 6 teams
    # will tend to become rather lengthy and should probably be executed
    # differently anyway.)
    #
    # Further reading
    # [1]: Scheduling Asynchronous Round-Robin Tournaments, W. Suksompong, arXiv:1804.04504 [math.CO]


    if len(teams) == 3:
        matchplan_indexes = matchplan_three_teams()
    elif len(teams) == 5:
        matchplan_indexes = matchplan_five_teams()
    else:
        matchplan_indexes = circle_method(len(teams))

    shuffled_teams = list(teams)
    random.shuffle(shuffled_teams)

    matchplan = []
    for match in matchplan_indexes:
        idx_a, idx_b = match
        a = shuffled_teams[idx_a]
        b = shuffled_teams[idx_b]
        matchplan.append((a, b))

    return matchplan

def matchplan_three_teams():
    # Without loss of generality, there should be exactly two possible matchplans with evenly distributed blue/red
    matchplans = [
        [(0, 1), (1, 2), (2, 0)],
        [(0, 1), (2, 0), (1, 2)]
    ]
    return random.choice(matchplans)

def matchplan_five_teams():
    # Without loss of generality, there are two possible matchplans for five teams (with the starting order undefined)
    # that don’t have the same team play two consecutive matches. (Assuming the teams are assigned randomly.)
    # (Source: pen&paper + brute force analysis)
    #
    # [{0, 1}, {2, 3}, {0, 4}, {1, 2}, {3, 4}, {0, 2}, {1, 3}, {2, 4}, {0, 3}, {1, 4}]
    # [{0, 1}, {2, 3}, {0, 4}, {1, 2}, {3, 4}, {0, 2}, {1, 4}, {0, 3}, {2, 4}, {1, 3}]

    matchplan = random.choice([
        [(0, 1), (2, 3), (0, 4), (1, 2), (3, 4), (0, 2), (1, 3), (2, 4), (0, 3), (1, 4)],
        [(0, 1), (2, 3), (0, 4), (1, 2), (3, 4), (0, 2), (1, 4), (0, 3), (2, 4), (1, 3)]
    ])

    # collect all matchplans that have an equal distribution of blue/red matches
    # TODO: It should be possible to reduce the possible combinations for this in advance.
    valid_matchplans = []
    for shuffle in itertools.product([0, 1], repeat=len(matchplan)):
        # for each possible order, make a histogram of each team playing blue
        # as we have five teams, we know that each team should play blue exactly twice
        if set(Counter(match[idx] for match, idx in zip(matchplan, shuffle)).values()) == {2}:
            valid_mp = [(match[idx], match[1 - idx]) for match, idx in zip(matchplan, shuffle)]
            valid_matchplans.append(valid_mp)

    return random.choice(valid_matchplans)

def circle_method(num_teams):
    return list(circle_method_gen(num_teams))

def circle_method_gen(num_teams):
    """ For the given number of teams, create a matchplan for a round-robin tournament using the circle method. """

    FILLER_TEAM = object()

    teams = list(range(num_teams))

    # add a dummy team if we have an odd number of teams
    # the matches will later be discarded
    if len(teams) % 2 != 0:
        teams.append(FILLER_TEAM)

    # choose an index that we keep fixed (must be in first or last column)
    #
    # entries are aligned clockwise
    #
    #  +---+---+---+---+
    #  | 0 | 1 | 2 | 3 |
    #  +---+---+---+---+
    #  | 4 | 5 | 6 | 7 |
    #  +---+---+---+---+
    #
    #  keep 0 fixed and rotate
    #
    #  +---+---+---+---+
    #  | 0 | 2 | 3 | 4 |
    #  +---+---+---+---+
    #  | 5 | 6 | 7 | 1 |
    #  +---+---+---+---+
    #
    # This gives us the starting matches: 04, 14, 26, 37, 50, 26, 37, 41, ...
    # Note that all matches of the fixed team need to be reversed in every second round

    # corner indexes
    if num_teams % 2 == 0:
        fixed_index = random.choice([0, len(teams) // 2 - 1, len(teams) // 2, len(teams) - 1])
    else:
        # for an odd number of teams, only select the column with the filler team
        fixed_index = random.choice([0, len(teams) - 1])

    flip_fixed = False
    for _iter in range(len(teams) - 1):
        teams = rotate_with_fixed(teams, fixed_index)

        # pair the matches
        for match_idx in range(len(teams) // 2):
            # pair a team from the top of the list with a team from the bottom
            top_index = match_idx
            bottom_index = len(teams) - match_idx - 1
            match = teams[top_index], teams[bottom_index]

            # the pairing with the fixed team needs to be flipped every other time
            if fixed_index in [top_index, bottom_index]:
                if flip_fixed:
                    match = teams[bottom_index], teams[top_index]
                flip_fixed = not flip_fixed

            if FILLER_TEAM in match:
                continue
            yield match

def rotate_with_fixed(lst: List, index: int) -> List:
    lst = list(lst)

    # remove index
    removed_team = lst.pop(index)
    # rotate
    lst = lst[1:] + lst[:1]
    # insert again
    lst.insert(index, removed_team)
    return lst



""" collecting the game state filter functions """
import random


def noiser(walls, shape, bot_position, enemy_positions, noise_radius=5, sight_distance=5, rnd=None):
    """Function to make bot positions noisy in a game state.

    Applies uniform noise in maze space. Noise will only be applied if the
    enemy bot is farther away than a certain threshold (`sight_distance`),
    which is Manhattan distance disregarding walls. A bot with distance of 1 in
    Manhattan space could still be much further away in maze distance.

    Distance to enemies measured in Manhattan space, disregarding walls. So, a
    bot distance of 1 in Manhattan space could still be much further away in
    maze distance.

    Given a `bot_position` and a list of `enemy_positions`, this function adds
    uniform noise in maze space to the enemy positions, but only if bot is
    farther away than sight_distance.

    The function returns a dictionary with entries `"enemy_positions"` (which
    holds the list of new enemy positions) and `"is_noisy"` (which is a list of
    bool, saying which enemy index is noisy).

    Functions needed
    ----------------

    altered_pos(bot_pos):
        return the noised new position of an enemy bot.

    manhattan_dist(a,b):
        returns a scalar

    Parameters
    ----------
    walls : set of (int, int)
    noise_radius : int, optional, default: 5
        the radius for the uniform noise
    sight_distance : int, optional, default: 5
        the distance at which noise is no longer applied.
    rnd : Random, optional
        the game’s random number generator (or None for an independent one)

    Returns
    -------
    dict { "enemy_positions": noised list of enemies, "is_noisy": list of bool }

    """

    # set the random state
    if rnd is None:
        rnd = random.Random()

    # store the noised positions
    noised_positions = [None] * len(enemy_positions)

    # store, if an enemy is noisy
    is_noisy = [None] * len(enemy_positions)

    for count, b in enumerate(enemy_positions):
        # Check that the distance between this bot and the enemy is larger
        # than `sight_distance`.
        cur_distance = manhattan_dist(bot_position, b)

        if cur_distance is None or cur_distance > sight_distance:
            # If so then alter the position of the enemy
            new_pos, noisy_flag = alter_pos(b, noise_radius, rnd, walls, shape)
            noised_positions[count] = new_pos
            is_noisy[count] = noisy_flag
        else:
            noised_positions[count] = b
            is_noisy[count] = False

    return { "enemy_positions": noised_positions, "is_noisy": is_noisy }


def alter_pos(bot_pos, noise_radius, rnd, walls, shape):
    """ alter the position """

    # get a list of possible positions
    x_min, x_max = bot_pos[0] - noise_radius, bot_pos[0] + noise_radius
    y_min, y_max = bot_pos[1] - noise_radius, bot_pos[1] + noise_radius

    # filter them so that we return no positions outside the maze
    if x_min < 0:
        x_min = 1
    if x_max >= shape[0]:
        x_max = shape[0] - 1
    if y_min < 0:
        y_min = 1
    if y_max >= shape[1]:
        y_max = shape[1] - 1

    possible_positions = [
        (i, j)
        for i in range(x_min, x_max + 1) # max + 1 since range is not inclusive at upper end
        for j in range(y_min, y_max + 1) # max + 1 since range is not inclusive at upper end
        if manhattan_dist((i, j), bot_pos) <= noise_radius
        and not (i, j) in walls # check that the bot won't returned as positioned on a wall square
    ]

    if len(possible_positions) < 1:
        # this should not happen
        # anyway. return the bot’s current position
        # TODO: Should probably raise?
        final_pos = bot_pos
        noisy = False
    elif len(possible_positions) == 1:
        final_pos = possible_positions[0]
        noisy = False
    else:
        # select a random position
        final_pos = rnd.choice(possible_positions)
        noisy = True

    # return the final_pos and a flag if it is noisy or not
    return (final_pos, noisy)

def in_homezone(position, team_id, shape):
    boundary = shape[0] / 2
    if team_id == 0:
        return position[0] < boundary
    elif team_id == 1:
        return position[0] >= boundary


def update_food_age(game_state, team, radius):
    # Only ghosts can cast a shadow
    ghosts = [
        bot for bot in game_state['bots'][team::2]
        if in_homezone(bot, team, game_state['shape'])
    ]
    food = game_state['food'][team]
    food_age = [dict(team_food_age) for team_food_age in game_state['food_age']]

    for pellet in food:
        if any(manhattan_dist(ghost, pellet) <= radius for ghost in ghosts):
            if pellet in food_age[team]:
                food_age[team][pellet] += 1
            else:
                food_age[team][pellet] = 1
        else:
            if pellet in food_age[team]:
                del food_age[team][pellet]

    return {'food_age': food_age}


def relocate_expired_food(game_state, team, radius, max_food_age=None):
    bots = game_state['bots'][team::2]
    enemy_bots = game_state['bots'][1-team::2]
    food = [set(team_food) for team_food in game_state['food']]
    food_age = [dict(team_food_age) for team_food_age in game_state['food_age']]
    width, height = game_state['shape']
    walls = game_state['walls']
    rnd = game_state['rnd']
    if max_food_age is None:
        max_food_age = game_state['max_food_age']

    # generate a set of possible positions to relocate food:
    #  - in the bot's homezone
    #  - not a wall
    #  - not on a already present food pellet
    #  - not on a bot
    #  - not on the border
    #  - not within the shadow of a bot
    home_width = width // 2
    left_most_x = home_width * team
    targets = { (x, y) for x in range(left_most_x, left_most_x+home_width) # this line and the next define the homezone
                       for y in range(height)
                       if (x not in (home_width, home_width - 1) # this excludes the border
                           and manhattan_dist(bots[0], (x, y)) > radius # this line and the next excludes the team's bots and their shadows
                           and manhattan_dist(bots[1], (x, y)) > radius )
              }
    targets = targets.difference(walls) # remove the walls
    targets = targets.difference(food[team]) # remove the team's food
    targets = targets.difference(enemy_bots) # remove the enemy bots
    # now convert to a list and sort, so that we have reproducibility (sets are unordered)
    targets = sorted(list(targets))
    for pellet in sorted(list(food[team])):
        # We move the pellet if it is in the food_age dict and exceeds the max_food_age
        if food_age[team].get(pellet, 0) > max_food_age:
            if not targets:
                # we have no free positions anymore, just let the food stay where it is
                # we do not update the age, so this pellet will get a chance to be
                # relocated at the next round
                continue
            # choose a new position at random
            new_pos = rnd.choice(targets)

            # remove the new pellet position from the list of possible targets for new pellets
            targets.remove(new_pos)

            # get rid of the old pellet
            food[team].remove(pellet)
            del food_age[team][pellet]

            # add the new pellet to food again
            # (starts with 0 food age, so we do not need to add it to the food_age dict)
            food[team].add(new_pos)

    return {'food' : food, 'food_age' : food_age}


def manhattan_dist(pos1, pos2):
    """ Manhattan distance between two points.

    Parameters
    ----------
    pos1 : tuple of (int, int)
        the first position
    pos2 : tuple of (int, int)
        the second position

    Returns
    -------
    manhattan_dist : int
        Manhattan distance between two points
    """
    return abs(pos1[0] - pos2[0]) + abs(pos1[1] - pos2[1])



import time

import pytest
import networkx

from pelita.layout import parse_layout, get_random_layout, initial_positions
from pelita.game import run_game, setup_game, play_turn

def stopping(bot, state):
    return bot.position

def randomBot(bot, state):
    legal = bot.legal_positions[:]
    return bot.random.choice(legal)

class TestStoppingTeam:
    @staticmethod
    def round_counting():
        storage_copy = {}
        def inner(bot, state):
            print(state)
            state[bot.turn] = state.get(bot.turn, 0) + 1
            storage_copy['rounds'] = state[bot.turn]
            print(state)
            return bot.position
        inner._storage = storage_copy
        return inner

    def test_stopping(self):
        test_layout = (
        """ ############
            #a#.by .# x#
            ############ """)

        round_counting = self.round_counting()
        team = [
            stopping,
            round_counting
        ]
        state = run_game(team, max_rounds=1, layout_dict=parse_layout(test_layout), allow_exceptions=True)
        assert state['bots'][0] == (1, 1)
        assert state['bots'][1] == (10, 1)
        assert round_counting._storage['rounds'] == 1

        round_counting = self.round_counting()
        team = [
            stopping,
            round_counting
        ]
        state = run_game(team, max_rounds=3, layout_dict=parse_layout(test_layout), allow_exceptions=True)
        assert state['bots'][0] == (1, 1)
        assert state['bots'][1] == (10, 1)
        assert round_counting._storage['rounds'] == 3

def test_track_and_kill_count():
    # for each team, we track whether they have been eaten at least once
    # and count the number of times they have been killed
    bot_states = {
        0: [{'track': [], 'eaten': False, 'times_killed': 0, 'deaths': 0},
            {'track': [], 'eaten': False, 'times_killed': 0, 'deaths': 0}],
        1: [{'track': [], 'eaten': False, 'times_killed': 0, 'deaths': 0},
            {'track': [], 'eaten': False, 'times_killed': 0, 'deaths': 0}]
    }
    def trackingBot(bot, state):
        turn = bot.turn
        other = bot.other

        # first move. get the state from the global cache
        if state == {}:
            team_idx = 0 if bot.is_blue else 1
            state.update(enumerate(bot_states[team_idx]))

        if bot.round == 1 and turn == 0:
            assert bot.track[0] == bot.position

        if bot.was_killed:
            state[turn]['eaten'] = True
            # if bot.deaths has increased from our last known value,
            # we add a kill
            # this avoids adding two kills as the eaten attribute could
            # be True in two consecutive turns
            if state[turn]['deaths'] != bot.deaths:
                state[turn]['times_killed'] += 1
            state[turn]['deaths'] = bot.deaths
        if other.was_killed:
            state[1 - turn]['eaten'] = True
            if state[1 - turn]['deaths'] != bot.other.deaths:
                state[1 - turn]['times_killed'] += 1
            state[1 - turn]['deaths'] = bot.other.deaths

        if bot.was_killed or not state[turn]['track']:
            state[turn]['track'] = [bot.position]
        if other.was_killed or not state[1 - turn]['track']:
            state[1 - turn]['track'] = [other.position]
        else:
            state[1 - turn]['track'].append(other.position)

        # The assertion is that the first position in bot.track
        # is always the respawn position.
        # However, in our test case, this will only happen, once
        # a bot has been eaten.
        if state[turn]['eaten']:
            assert bot.track[0] == bot._initial_position
        assert bot.track == state[turn]['track'] # bot.round * 2 + 1 + turn
        assert bot.track[-1] == bot.position
        # just move randomly. hopefully, this means some bots will be killed
        return randomBot(bot, state)

    layout = """
    ##########
    # ab .y x#
    # ########
    #.##. .###
    ##########
    """
    team = [
        trackingBot,
        trackingBot
    ]
    # We play 600 rounds as we rely on some randomness in our assertions
    state = setup_game(team, max_rounds=600, allow_camping=True, layout_dict=parse_layout(layout))
    while not state['gameover']:
        # Check that our count is consistent with what the game thinks
        # for the current and previous bot, we have to subtract the deaths that have just respawned
        # as they have not been passed to the bot yet
        # therefore, we only update old_deaths for the current team

        if state['turn'] is not None:
            old_deaths[state['turn']] = state['deaths'][state['turn']]

        old_deaths = state['deaths'][:]
        state = play_turn(state)
        deaths = state['deaths'][:]

        team = state['turn'] % 2

        # The current bot knows about its deaths, *unless* it made suicide,
        # so we have to subtract 1 if bot_was_killed is True
        suicide_correction = [0] * 4
        suicide_correction[state['turn']] = 1 if state['bot_was_killed'][state['turn']] else 0

        # The other team knows about its deaths, *unless* one of the bots got eaten
        # just now, or the previous bot made suicide
        other_team_correction = [0] * 4
        for idx in range(1 - team, 4, 2):
            if old_deaths[idx] != deaths[idx]:
                other_team_correction[idx] = 1

        # suicide
        prev_idx = state['turn'] - 1
        if old_deaths[prev_idx] == deaths[prev_idx] and state['bot_was_killed'][prev_idx]:
            other_team_correction[prev_idx] = 1

        assert bot_states[0][0]['times_killed'] == deaths[0] - suicide_correction[0] - other_team_correction[0]
        assert bot_states[1][0]['times_killed'] == deaths[1] - suicide_correction[1] - other_team_correction[1]
        assert bot_states[0][1]['times_killed'] == deaths[2] - suicide_correction[2] - other_team_correction[2]
        assert bot_states[1][1]['times_killed'] == deaths[3] - suicide_correction[3] - other_team_correction[3]

        # assertions might have been caught in run_game
        # check that all is good
        assert state['fatal_errors'] == [[], []]

    # check that the game has run at all
    assert state['round'] >= 1
    # check that someone has been killed, or the whole test is not doing anything
    assert sum(state['deaths']) > 0
    # check that each single bot has been eaten, or we are not testing the full range of possibilities
    assert all(state['deaths'])


@pytest.mark.parametrize('bot_to_move', range(4))
def test_eaten_flag_kill(bot_to_move):
    """ Test that the eaten flag is set correctly in kill situations. """
    layout = """
    ########
    #  xa  #
    #  yb  #
    #......#
    ########
    """
    def move(bot, state):
        x, y = bot.position
        new_pos = bot.position
        if bot_to_move == 0:
            # we move in the first round as blue team in turn == 0
            if bot.round == 1 and bot.is_blue and bot.turn == 0:
                new_pos = (x - 1, y)
            # The red team should notice immediately
            if bot.round == 1 and not bot.is_blue and bot.turn == 0:
                assert bot.was_killed
                assert bot.other.was_killed is False
            # as bot 1 has been moved, the eaten flag will be reset in all other cases
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False
        if bot_to_move == 1:
            # we move in the first round as red team
            if bot.round == 1 and not bot.is_blue and bot.turn == 0:
                new_pos = (x + 1, y)
            # The other team should notice immediately that its other bot (#0) has been eaten
            if bot.round == 1 and bot.is_blue and bot.turn == 1:
                assert bot.was_killed is False
                assert bot.other.was_killed
            # When bot 0 moves, the flag is still set
            elif bot.round == 2 and bot.is_blue and bot.turn == 0:
                assert bot.was_killed
                assert bot.other.was_killed is False
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False
        if bot_to_move == 2:
            # we move in the first round as blue team in turn == 1
            if bot.round == 1 and bot.is_blue and bot.turn == 1:
                new_pos = (x - 1, y)
            # The red team should notice immediately
            if bot.round == 1 and not bot.is_blue and bot.turn == 1:
                assert bot.was_killed
                assert bot.other.was_killed is False
            # as bot 2 has been moved, the eaten flag will be reset in all other cases
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False
        if bot_to_move == 3:
            # we move in the first round as red team in turn == 1
            if bot.round == 1 and not bot.is_blue and bot.turn == 1:
                new_pos = (x + 1, y)
            # The blue team should notice immediately (in round == 2!) that bot 1 has been eaten
            if bot.round == 2 and bot.is_blue and bot.turn == 0:
                assert bot.was_killed is False
                assert bot.other.was_killed
            # When bot 1 moves, the flag is still set
            elif bot.round == 2 and bot.is_blue and bot.turn == 1:
                assert bot.was_killed
                assert bot.other.was_killed is False
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False

        # otherwise return current position
        return new_pos
    state = run_game([move, move], max_rounds=3, layout_dict=parse_layout(layout), allow_exceptions=True)
    # assertions might have been caught in run_game
    # check that all is good
    assert state['fatal_errors'] == [[], []]


@pytest.mark.parametrize("bot_to_move", range(4))
def test_eaten_flag_suicide(bot_to_move):
    """ Test that the eaten flag is set correctly in suicide situations. """
    layout = """
    ########
    #  ax  #
    #  by  #
    #......#
    ########
    """
    def move(bot, state):
        x, y = bot.position
        new_pos = bot.position
        if bot_to_move == 0:
            # we move in the first round as blue team in turn == 0
            if bot.round == 1 and bot.is_blue and bot.turn == 0:
                new_pos = (x + 1, y)
            # our other bot should notice it in our next turn
            if bot.round == 1 and bot.is_blue and bot.turn == 1:
                assert bot.was_killed is False
                assert bot.other.was_killed
            # bot 0 will notice it in the next round
            elif bot.round == 2 and bot.is_blue and bot.turn == 0:
                assert bot.was_killed
                assert bot.other.was_killed is False
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False
        if bot_to_move == 1:
            # we move in the first round as red team
            if bot.round == 1 and not bot.is_blue and bot.turn == 0:
                new_pos = (x - 1, y)
            # we should notice in our next turn
            if bot.round == 1 and not bot.is_blue and bot.turn == 1:
                assert bot.was_killed is False
                assert bot.other.was_killed
            # bot 1 will notice it in the next round
            elif bot.round == 2 and not bot.is_blue and bot.turn == 0:
                assert bot.was_killed
                assert bot.other.was_killed is False
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False
        if bot_to_move == 2:
            # we move in the first round as blue team in turn == 1
            if bot.round == 1 and bot.is_blue and bot.turn == 1:
                new_pos = (x + 1, y)
            # we should notice in our next turn (next round!)
            if bot.round == 2 and bot.is_blue and bot.turn == 0:
                assert bot.was_killed is False
                assert bot.other.was_killed
            # bot 2 will notice it in the next round as well
            elif bot.round == 2 and bot.is_blue and bot.turn == 1:
                assert bot.was_killed
                assert bot.other.was_killed is False
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False
        if bot_to_move == 3:
            # we move in the first round as red team in turn == 1
            if bot.round == 1 and not bot.is_blue and bot.turn == 1:
                new_pos = (x - 1, y)
            # we should notice in our next turn (next round!)
            if bot.round == 2 and not bot.is_blue and bot.turn == 0:
                assert bot.was_killed is False
                assert bot.other.was_killed
            # bot 3 will notice it in the next round as well
            elif bot.round == 2 and not bot.is_blue and bot.turn == 1:
                assert bot.was_killed
                assert bot.other.was_killed is False
            else:
                assert bot.was_killed is False
                assert bot.other.was_killed is False

        # otherwise return current position
        return new_pos
    state = run_game([move, move], max_rounds=3, layout_dict=parse_layout(layout), allow_exceptions=True)
    # assertions might have been caught in run_game
    # check that all is good
    assert state['fatal_errors'] == [[], []]


@pytest.mark.parametrize('_n_test', range(10))
def test_initial_position(_n_test):
    """ Test that out test team receives the correct initial positions."""
    layout_name, layout_string = get_random_layout()
    l = parse_layout(layout_string)
    initial_pos = initial_positions(l['walls'], l['shape'])

    def move(bot, state):
        if bot.is_blue and bot.turn == 0:
            assert bot._initial_position == initial_pos[0]
            assert bot.other._initial_position == initial_pos[2]
            assert bot.enemy[0]._initial_position == initial_pos[1]
            assert bot.enemy[1]._initial_position == initial_pos[3]
        if bot.is_blue and bot.turn == 1:
            assert bot._initial_position == initial_pos[2]
            assert bot.other._initial_position == initial_pos[0]
            assert bot.enemy[0]._initial_position == initial_pos[1]
            assert bot.enemy[1]._initial_position == initial_pos[3]
        if not bot.is_blue and bot.turn == 0:
            assert bot._initial_position == initial_pos[1]
            assert bot.other._initial_position == initial_pos[3]
            assert bot.enemy[0]._initial_position == initial_pos[0]
            assert bot.enemy[1]._initial_position == initial_pos[2]
        if not bot.is_blue and bot.turn == 1:
            assert bot._initial_position == initial_pos[3]
            assert bot.other._initial_position == initial_pos[1]
            assert bot.enemy[0]._initial_position == initial_pos[0]
            assert bot.enemy[1]._initial_position == initial_pos[2]
        return randomBot(bot, state)

    state = run_game([move, move], max_rounds=3, layout_dict=l)
    # assertions might have been caught in run_game
    # check that all is good
    assert state['fatal_errors'] == [[], []]


def test_bot_attributes():
    test_layout = """
        ##################
        #.#... .##.     y#
        # #a#  .  .### #x#
        # ####.   .      #
        #      .   .#### #
        # # ###.  .  # # #
        #b     .##. ...#.#
        ##################
    """

    parsed = parse_layout(test_layout)
    width = max(parsed['walls'])[0] + 1
    height = max(parsed['walls'])[1] + 1
    homezones = [[], []]
    homezones[0] = [(x, y)
                    for x in range(0, width // 2)
                    for y in range(0, height)
                    if (x, y) not in parsed['walls']]
    homezones[1] = [(x, y)
                    for x in range(width // 2, width)
                    for y in range(0, height)
                    if (x, y) not in parsed['walls']]

    assert set(homezones[0]) & set(homezones[1]) == set()

    def asserting_team(bot, state):
        assert bot.homezone == bot.other.homezone
        assert bot.walls == parsed['walls']

        assert bot.walls == tuple(sorted(bot.walls))
        assert bot.homezone == tuple(sorted(bot.homezone))

        if bot.is_blue:
            assert set(bot.homezone) == set(homezones[0])
            assert set(bot.enemy[0].homezone) == set(homezones[1])
            assert set(bot.shaded_food) == set([(1, 1), (3, 1), (4, 1), (5, 1)])
            assert set(bot.other.shaded_food) == set([(1, 1), (3, 1), (4, 1), (5, 1)])
            assert set(bot.enemy[0].shaded_food) == set()
            assert set(bot.enemy[1].shaded_food) == set()
        else:
            assert set(bot.homezone) == set(homezones[1])
            assert set(bot.enemy[0].homezone) == set(homezones[0])
            assert set(bot.shaded_food) == set()
            assert set(bot.other.shaded_food) == set()
            assert set(bot.enemy[0].shaded_food) == set()
            assert set(bot.enemy[1].shaded_food) == set()
        return bot.position

    state = run_game([asserting_team, asserting_team], max_rounds=1, layout_dict=parsed)
    # assertions might have been caught in run_game
    # check that all is good
    assert state['errors'] == [{}, {}]
    assert state['fatal_errors'] == [[], []]

def test_bot_graph():
    layout = """
    ########
    #  ax  #
    #  by  #
    #......#
    ########
    """
    def rough_bot(bot, state):
        # check that we have this node in the graph at the beginning
        assert (1, 1) in bot.graph
        # check that we can't remove a node
        with pytest.raises(networkx.NetworkXError):
            bot.graph.remove_node((1, 1))
        # check that we can't add a node
        with pytest.raises(networkx.NetworkXError):
            bot.graph.add_node((100, 100))
        # check that I can create my own copy of the graph and that this copy is
        # writable
        my_graph = bot.graph.copy()
        my_graph.remove_node((1, 1))
        assert (1, 1) not in my_graph
        return bot.position

    state  = run_game([rough_bot, stopping], max_rounds=1, layout_dict=parse_layout(layout))
    # assertions might have been caught in run_game
    # check that all is good
    assert state['errors'] == [{}, {}]
    assert state['fatal_errors'] == [[], []]

def test_bot_graph_is_half_mutable():
    # Test that a bot can change the weights of the graph
    # Changing the weights affects the future self of the bot
    # Changing the weights does affect the other bot in the team
    # Changing the weights does not affect the other team
    layout = """
    ########
    #  ax  #
    #  by  #
    #......#
    ########
    """

    observer = []

    def blue(bot, state):
        if bot.turn == 0 and bot.round == 1:
            assert bot.graph[1, 1][1, 2].get('weight') is None
            bot.graph[1, 1][1, 2]['weight'] = 1
            observer.append('B')
        else:
            assert bot.graph[1, 1][1, 2].get('weight') == 1
            observer.append('b')

        return bot.position

    def red(bot, state):
        if bot.turn == 0 and bot.round == 1:
            assert bot.graph[1, 1][1, 2].get('weight') is None
            bot.graph[1, 1][1, 2]['weight'] = 2
            observer.append('R')
        else:
            assert bot.graph[1, 1][1, 2].get('weight') == 2
            observer.append('r')

        return bot.position

    state  = run_game([blue, red], max_rounds=2, layout_dict=parse_layout(layout))
    # assertions might have been caught in run_game
    # check that all is good
    assert state['errors'] == [{}, {}]
    assert state['fatal_errors'] == [[], []]
    assert "".join(observer) == 'BRbrbrbr'


def test_team_names():
    test_layout = (
    """ ##################
        #a#.  .  # .     #
        #b#####    #####x#
        #     . #  .  .#y#
        ################## """)

    def team_pattern(fn):
        # The pattern for a local team.
        return f'local-team ({fn})'

    def team_1(bot, state):
        assert bot.team_name == team_pattern('team_1')
        assert bot.other.team_name == team_pattern('team_1')
        assert bot.enemy[0].team_name == team_pattern('team_2')
        assert bot.enemy[1].team_name == team_pattern('team_2')
        return bot.position

    def team_2(bot, state):
        assert bot.team_name == team_pattern('team_2')
        assert bot.other.team_name == team_pattern('team_2')
        assert bot.enemy[0].team_name == team_pattern('team_1')
        assert bot.enemy[1].team_name == team_pattern('team_1')
        return bot.position

    state = setup_game([team_1, team_2], layout_dict=parse_layout(test_layout), max_rounds=3)
    assert state['team_names'] == [team_pattern('team_1'), team_pattern('team_2')]

    state = play_turn(state)
    # check that player did not fail
    assert state['errors'] == [{}, {}]
    assert state['fatal_errors'] == [[], []]

    state = play_turn(state)
    # check that player did not fail
    assert state['errors'] == [{}, {}]
    assert state['fatal_errors'] == [[], []]


def test_team_time():
    test_layout = (
    """ ##################
        #a#.  .  # .     #
        #b#####    #####x#
        #     . #  .  .#y#
        ################## """)

    def team_pattern(fn):
        # The pattern for a local team.
        return f'local-team ({fn})'

    # NB: time.monotonic() makes huge leaps on windows,
    # we must therefore test with >= instead of just >
    # although this is a little stupid

    check = [None, None]

    def team_1(bot, state):
        if bot.round == 1 and bot.turn == 0:
            assert bot.team_time == 0
            assert bot.enemy[0].team_time == 0
        else:
            assert bot.team_time >= 0

        # we fake the time in round 2!!
        if bot.round == 2 and bot.turn == 0:
            assert bot.team_time == 1.0
            assert bot.enemy[0].team_time == 2.0

        if bot.round == 2 and bot.turn == 1:
            assert bot.team_time >= 1.0
            assert bot.enemy[0].team_time >= 2.0

        check[0] = bot.team_time

        return bot.position

    def team_2(bot, state):
        if bot.round == 1 and bot.turn == 0:
            assert bot.team_time == 0
            assert bot.enemy[0].team_time >= 0
        else:
            assert bot.team_time >= 0

        # we fake the time in round 2!!
        if bot.round == 2 and bot.turn == 0:
            assert bot.team_time == 2.0
            assert bot.enemy[0].team_time >= 1.0

        if bot.round == 2 and bot.turn == 1:
            assert bot.team_time >= 2.0
            assert bot.enemy[0].team_time >= 1.0

        check[1] = bot.team_time

        return bot.position

    state = setup_game([team_1, team_2], layout_dict=parse_layout(test_layout), max_rounds=3, allow_exceptions=True)

    state = play_turn(state)
    assert state['team_time'][0] >= 0
    assert state['team_time'][1] == 0

    state = play_turn(state)
    assert state['team_time'][0] >= 0
    assert state['team_time'][1] >= 0

    # time before bots get updated
    old_time = list(state['team_time'])

    state = play_turn(state)
    state = play_turn(state)

    assert check == old_time

    # Round 2. We fake the time so we do not have to sleep!
    state['team_time'] = [1.0, 2.0]

    state = play_turn(state)
    state = play_turn(state)

    # time before bots get updated
    old_time = list(state['team_time'])

    state = play_turn(state)
    state = play_turn(state)

    assert check == old_time

    # check that player did not fail
    assert state['errors'] == [{}, {}]
    assert state['fatal_errors'] == [[], []]
    assert state['team_time'][0] >= 1.0
    assert state['team_time'][1] >= 2.0


def test_bot_str_repr():
    test_layout = """
        ##################
        #.#... .##.     y#
        # # #  .  .### #x#
        # ####.   .      #
        #      .   .#### #
        #a# ###.  .  # # #
        #b     .##. ...#.#
        ##################
    """

    parsed = parse_layout(test_layout)

    def asserting_team(bot, state):
        bot_str = str(bot).split('\n')
        if bot.is_blue and bot.round == 1:
            assert bot_str[0] == "local-team (asserting_team) (you) vs local-team (asserting_team)."
            assert bot_str[1] == f"Playing on blue side. Current turn: {bot.turn}. Bot: {bot.char}. Round: 1, score: 0:0. timeouts: 0:0"
        elif not bot.is_blue and bot.round == 1:
            assert bot_str[0] == "local-team (asserting_team) vs local-team (asserting_team) (you)."
            assert bot_str[1] == f"Playing on red side. Current turn: {bot.turn}. Bot: {bot.char}. Round: 1, score: 0:0. timeouts: 0:0"
        else:
            assert False, "Should never be here."

        return bot.position

    state = run_game([asserting_team, asserting_team], max_rounds=1, layout_dict=parsed,
                     allow_exceptions=True)
    # assertions might have been caught in run_game
    # check that all is good
    assert state['fatal_errors'] == [[], []]


def test_bot_html_repr():
    test_layout = """
        ##################
        #.#... .##.     y#
        # # #  .  .### #x#
        # ####.   .      #
        #      .   .#### #
        #a# ###.  .  # # #
        #b     .##. ...#.#
        ##################
    """

    parsed = parse_layout(test_layout)

    def asserting_team(bot, state):
        # Not a full-fledged test at this time. We mainly want to catch API changes for now.

        bot_str = bot._repr_html_()
        assert len(bot_str)

        return bot.position

    state = run_game([asserting_team, asserting_team], max_rounds=1, layout_dict=parsed,
                     allow_exceptions=True)
    # assertions might have been caught in run_game
    # check that all is good
    assert state['fatal_errors'] == [[], []]




""" The observers. """

import json
import logging
import sys

import zmq
from rich.console import Console
from rich.progress import (BarColumn, MofNCompleteColumn, Progress,
                           SpinnerColumn, TextColumn, TimeElapsedColumn)

from . import layout
from .network import SetEncoder

_logger = logging.getLogger(__name__)
_mswindows = (sys.platform == "win32")

# Only highlight explicit markup
pprint = Console(highlight=False).print

class ProgressViewer:
    def __init__(self) -> None:
        self.progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
        )
        self.task = self.progress.add_task('[Pelita]', total=None)

    def show_state(self, game_state):
        score = game_state["score"]
        round_index = game_state["round"]
        if round_index is None:
            return
        game_time = game_state["max_rounds"]

        turn = game_state["turn"]
        if turn is not None:
            bot_idx = turn // 2
            if turn % 2 == 0:
                bot_sign = f'[blue]Blue Team, Bot {bot_idx}[/]'
            elif turn % 2 == 1:
                bot_sign = f'[red]Red Team,  Bot {bot_idx}[/]'
        else:
            bot_sign = ' '

        score = ":".join(str(s) for s in score)
        status = f"[{bot_sign}] [{score}]"

        self.progress.update(self.task, description=status, total=game_time,
                             completed=round_index, refresh=True)
        self.progress.start()

        if game_state["gameover"]:
            state = {}
            state.update(game_state)
            del state['walls']
            del state['food']

            self.progress.stop()
            self.progress.console.clear_live()

            print()
            print(f"Final state: {state}")


class AsciiViewer:
    """ A viewer that dumps ASCII charts on stdout. """

    def color_bots(self, layout_str):
        out_str = layout_str
        # we have to do the replace in two steps to avoid
        # that the ‘b’ in ‘[blue]’ gets replaced with itself
        for turn, char in layout.BOT_I2N.items():
            team = turn % 2
            col = '[B]' if team == 0 else '[R]'
            out_str = out_str.replace(char, f'{col}{char}[/]')

        out_str = out_str.replace('B', 'blue').replace('R', 'red')

        return out_str

    def show_state(self, game_state):
        uni_str = layout.layout_as_str(walls=game_state['walls'],
                                       food=game_state['food'],
                                       bots=game_state['bots'])

        # color bots
        uni_str = self.color_bots(uni_str)
        # Everything that we print explicitly is removed from the state dict.
        state = {}
        state.update(game_state)
        # for death and kills just leave a summary
        state['blue deaths'] = state['deaths'][::2]
        state['blue kills'] = state['kills'][::2]
        state['red deaths'] = state['deaths'][1::2]
        state['red kills'] = state['kills'][1::2]
        del state['kills']
        del state['deaths']
        del state['walls']
        del state['food']
        del state['bots']
        del state['round']
        del state['turn']
        del state['score']

        turn = game_state["turn"]
        if turn is not None:
            team = 'Blue' if turn % 2 == 0 else 'Red'
            bot_idx = turn // 2
            bot_name = layout.BOT_I2N[bot_idx]
        else:
            team = '–'
            bot_name = '–'
        round=game_state["round"]
        s0=game_state["score"][0]
        s1=game_state["score"][1]
        state=state
        universe=uni_str
        length = len(universe.splitlines()[0])
        info = (
            f"Round: {round!r} | Team: {team} | Bot: {bot_name} | Score {s0}:{s1}\n"
            f"Game State: {state!r}\n"
            f"\n"
        )

        print(info)
        pprint(universe)
        print("–"*length)
        if state.get("gameover"):
            if state["whowins"] == 2:
                pprint("Game Over: Draw.")
            else:
                winner = game_state["team_names"][state["whowins"]]
                pprint(f"Game Over: Team: '{winner}' wins!")


class ReplyToViewer:
    """ A viewer which dumps to a given stream.
    """
    def __init__(self, reply_to):
        ctx = zmq.Context()
        self.sock = ctx.socket(zmq.PAIR)

        # Wait max linger ms for a socket to connect
        # before giving up.
        self.sock.linger = 1000

        self.sock.connect(reply_to)
        _logger.debug(f"Connecting zmq.PAIR to {reply_to}")

        self.pollout = zmq.Poller()
        self.pollout.register(self.sock, zmq.POLLOUT)

    def _send(self, message):
        socks = dict(self.pollout.poll(300))
        if socks.get(self.sock) == zmq.POLLOUT:
            as_json = json.dumps(message, cls=SetEncoder)
            self.sock.send_unicode(as_json, flags=zmq.NOBLOCK)

    def show_state(self, game_state):
        self._send(game_state)


class ReplayWriter:
    """ A viewer which dumps to a given stream.
    """
    def __init__(self, stream):
        self.stream = stream

    def _send(self, message):
        as_json = json.dumps(message, cls=SetEncoder)
        self.stream.write(as_json)
        # We use 0x04 (EOT) as a separator between the events.
        # The additional newline is for improved readability
        # and should be ignored by the Python json reader.
        self.stream.write("\x04\n")
        self.stream.flush()

    def show_state(self, game_state):
        self._send(game_state)


class ResultPrinter:
    def show_state(self, state):
        if (state['turn'] is None and
            state['round'] is None):
            # TODO: We must ensure that this is only printed once
            self.print_team_names(state['team_names'])

        if state["gameover"]:
            self.print_possible_winner(state)

    def print_team_names(self, team_names):
        # TODO: The team_spec is missing in the state.
        # Should we print it here?
        pie = '' if _mswindows else 'ᗧ'

        for col, team_name in zip(['blue', 'red'], team_names):
            if team_name is not None:
                pprint(f"[bright_{col}]{pie}[/bright_{col}] {col} team: '{team_name}'")

    def print_possible_winner(self, state):
        """ Checks the game state for a winner.

        This is needed for pelita.scripts parsing the output.
        """
        winning_team = state.get("whowins")
        if state['round'] == 1:
            num_rounds = "1 round"
        else:
            num_rounds = f"{state['round']} rounds"
        if winning_team in (0, 1):
            winner = state['team_names'][winning_team]
            loser = state['team_names'][1 - winning_team]
            winner_score = state['score'][winning_team]
            loser_score = state['score'][1 - winning_team]
            msg = f"Finished after {num_rounds}. '{winner}' won over '{loser}'. ({winner_score}:{loser_score})"
        elif winning_team == 2:
            t1, t2 = state['team_names']
            s1, s2 = state['score']
            msg = f"Finished after {num_rounds}. '{t1}' and '{t2}' had a draw. ({s1}:{s2})"
        else:
            return

        # We must flush, else our forceful stopping of Tk
        # won't let us pipe it.
        print(msg, flush=True)




import json
import logging
import sys
import time
from urllib.parse import urlparse
import uuid

import zmq

_logger = logging.getLogger(__name__)

# 41736 is the word PELI(T)A when read upside down in reverse without glasses
# The missing T stands for tcp
PELITA_PORT = 41736

## Pelita network data structures

# ControlRequest
# {__action__}

# ViewerUpdate
# {__action__, __data__}

# Request
# {__uuid__, __action__, __data__}

# Reply
# {__uuid__, __return__}

# Error
# {__uuid__, __error__, __error_msg__}


class ZMQUnreachablePeer(Exception):
    """ Raised when ZMQ cannot send a message (connection may have been lost). """


class ZMQReplyTimeout(Exception):
    """ Is raised when an ZMQ socket does not answer in time. """


class ZMQClientError(Exception):
    """ Used to propagate errors from the client.
    Raised when the zmq connection receives an __error__ message. """
    def __init__(self, message, error_type, *args):
        self.message = message
        self.error_type = error_type
        super().__init__(message, error_type, *args)


#: The timeout to use during sending
DEAD_CONNECTION_TIMEOUT = 3.0


def bind_socket(socket: zmq.Socket, address, option_hint=None):
    parsed_address = urlparse(address)

    # NB: We cannot use parsed_address.geturl() to generate a nice url for zmq
    # as this will eat the empty hostname in a file path and zmq does not like that.
    # file:///tmp/a -> file:/tmp/a

    try:
        if parsed_address.scheme == 'tcp' and parsed_address.port is None:
            port = socket.bind_to_random_port(address)
            return f'tcp://{parsed_address.hostname}:{port}'
        else:
            socket.bind(address)
            return address

    except (zmq.ZMQError, zmq.ZMQBindError) as e:
        print('error binding to address %s: %s' % (address, e), file=sys.stderr)
        if option_hint:
            print('use %s <address> to specify a different port' %\
                (option_hint,), file=sys.stderr)
        raise


class SetEncoder(json.JSONEncoder):
   def default(self, obj):
      if isinstance(obj, set):
         return list(obj)
      return json.JSONEncoder.default(self, obj)


def json_default_handler(o):
    """ Pythons built-in json handler has problems converting numpy.in64
    to json. By adding this method as a default= to json.dumps, we can
    tell Python what to do.
    """
    try:
        import numpy as np
        if isinstance(o, np.integer):
            return int(o)
    except ImportError:
        pass
    # we don’t know the type: raise a Type error
    raise TypeError("Cannot convert %r of type %s to json" % (o, type(o)))


class ZMQConnection:
    """ This class is supposed to ease request–reply connections
    through a zmq socket. It does so by attaching a uuid to each
    request. It will only accept a reply if this also includes
    this uuid. All other incoming messages will be discarded.

    Please note the following:
      * This class is not thread-safe!
      * There can only be one request at a time. Only the reply for
        the most recent request (= uuid) will be received. Non-matching
        uuids are discarded.
      * There is no storage of messages.

    Parameters
    ----------
    socket : zmq socket
        The zmq socket of this connection

    Attributes
    ----------
    socket : zmq socket
        The zmq socket of this connection
    pollin : zmq poller
        Poller for incoming connections
    pollout : zmq poller
        Poller for outgoing connections
    """
    def __init__(self, socket):
        self.socket = socket

        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.setsockopt(zmq.AFFINITY, 1)
        #self.setsockopt(zmq.RCVTIMEO, 2000)

        self.pollin = zmq.Poller()
        self.pollin.register(socket, zmq.POLLIN)
        self.pollout = zmq.Poller()
        self.pollout.register(socket, zmq.POLLOUT)

    def send(self, action, data, timeout=None):
        """ Sends a message or request `action`
        and attached data to the socket and returns the
        message id that is needed to receive the reply.
        """

        if timeout is None:
            timeout = DEAD_CONNECTION_TIMEOUT

        msg_id = str(uuid.uuid4())
        _logger.debug("---> %r [%s]", action, msg_id)

        # Check before sending that the socket can receive
        socks = dict(self.pollout.poll(timeout * 1000))
        if socks.get(self.socket) == zmq.POLLOUT:
            # I think we need to set NOBLOCK here, else we may run into a
            # race condition if a connection was closed between poll and send.
            # NOBLOCK should raise, so we can catch that
            message_obj = {"__uuid__": msg_id, "__action__": action, "__data__": data}
            json_message = json.dumps(message_obj, cls=SetEncoder)
            try:
                self.socket.send_unicode(json_message, flags=zmq.NOBLOCK)
            except zmq.ZMQError as e:
                _logger.info("Could not send message. Assume socket is unavailable. %r", e)
                raise ZMQUnreachablePeer()
        else:
            raise ZMQUnreachablePeer()
        return msg_id

    def _recv(self):
        """ Receive the next message on the socket.

        Returns
        -------
        (msg_id, reply)
            The message id and its data.

        Raises
        ------
        ZMQReplyTimeout
            if the message cannot be parsed from JSON
        ZMQClientError
            if an error message is returned
        """
        json_message = self.socket.recv_unicode()
        try:
            py_obj = json.loads(json_message)
        except ValueError:
            _logger.warning('Received non-json message from self. Triggering a timeout.')
            raise ZMQReplyTimeout()

        try:
            error_type = py_obj['__error__']
            error_message = py_obj.get('__error_msg__', '')
            _logger.warning(f'Received error reply ({error_type}): {error_message}. Closing socket.')
            self.socket.close()
            raise ZMQClientError(error_message, error_type)
        except KeyError:
            pass

        try:
            msg_id = py_obj["__uuid__"]
        except KeyError:
            msg_id = None
            _logger.warning('__uuid__ missing in message.')

        msg_return = py_obj.get("__return__")

        _logger.debug("<--- %r [%s]", msg_return, msg_id)
        return msg_id, msg_return

    def recv_timeout(self, expected_id, timeout):
        """ Waits `timeout` seconds for a reply with msg_id `expected_id`.

        Returns
        -------
        reply
            The reply for the `expected_id`

        Raises
        ------
        ZMQReplyTimeout
            if the message cannot be parsed from JSON
            if the message was not received in time
        ZMQConnectionError
            if an error message is returned
        """
        # special case for no timeout
        # just loop until we receive the correct reply
        if timeout is None:
            while True:
                msg_id, reply = self._recv()
                if msg_id == expected_id:
                    return reply

        # normal timeout handling

        time_now = time.monotonic()
        # calculate until when it may take
        # NB: When rewriting this code,
        # ensure that the case timeout=0
        # can still be handled
        timeout_until = time_now + timeout

        while time_now < timeout_until:
            time_left = timeout_until - time_now

            socks = dict(self.pollin.poll(time_left * 1000)) # poll needs milliseconds
            if socks.get(self.socket) == zmq.POLLIN:
                msg_id, reply = self._recv()

                # check, if it is the correct reply and return
                if msg_id == expected_id:
                    return reply

                else:
                    # We received a message with the wrong id.
                    # Reset the current time and try again.
                    time_now = time.monotonic()
                    continue
            else:
                # poll timed out
                # answer did not arrive in time
                break

        raise ZMQReplyTimeout()

    def __repr__(self):
        return "ZMQConnection(%r)" % self.socket

class ZMQPublisher:
    """ Sets up a simple Publisher which sends all viewed events
    over a zmq connection.

    Parameters
    ----------
    address : string
        The address which the publisher binds.
    """
    def __init__(self, address):
        self.address = address
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUB)
        self.socket_addr = bind_socket(self.socket, self.address, '--publish')
        _logger.debug("Bound zmq.PUB to {}".format(self.socket_addr))

    def _send(self, action, data):
        info = {'round': data['round'], 'turn': data['turn']}
        if data['gameover']:
            info['gameover'] = True
        _logger.debug(f"--#> [{action}] %r", info)
        message = {"__action__": action, "__data__": data}
        as_json = json.dumps(message, cls=SetEncoder)
        self.socket.send_unicode(as_json)

    def show_state(self, game_state):
        self._send(action="observe", data=game_state)


class Controller:
    def __init__(self, address='tcp://127.0.0.1', zmq_context=None):
        self.address = address
        if zmq_context:
            self.context = zmq_context
        else:
            self.context = zmq.Context()
        # We use a ROUTER which we bind.
        # This means other DEALERs can connect and
        # each one can take over control.
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket_addr = bind_socket(self.socket, self.address)
        self.pollin = zmq.Poller()
        self.pollin.register(self.socket, zmq.POLLIN)
        _logger.debug("Bound zmq.ROUTER to {}".format(self.socket_addr))

    def await_action(self, await_action, timeout=None, accept_exit=True):
        """ Waits `timeout` seconds to receive an action. """
        t_start = time.monotonic()
        if timeout is None:
            t_end = float("inf")
        else:
            t_end = t_start + timeout

        while time.monotonic() < t_end:

            # TODO: Proper time left handling
            timeoutmillis = timeout * 1000 if timeout is not None else None

            sock = dict(self.pollin.poll(timeoutmillis)) # poll needs milliseconds
            if sock.get(self.socket) == zmq.POLLIN:
                try:
                    sender, msg = self.socket.recv_multipart()
                    msg = json.loads(msg)
                    action = msg['__action__']
                    _logger.debug('<=== %r from %r', action, sender)
                except ValueError:
                    _logger.warning('Could not deserialize message from %r', sender)
                    continue
                except (TypeError, KeyError):
                    _logger.warning('No action in message from %r', sender)
                    continue

                expected_actions = [await_action]
                if accept_exit:
                    expected_actions.append('exit')
                if action in expected_actions:
                    return action
                _logger.warning('Unexpected action %r. (Expected: %s) Ignoring.', action, ", ".join(expected_actions))
                continue


    def recv_start(self, timeout=None):
        """ Waits `timeout` seconds for start message.

        Returns `True`, when the message arrives, `False` when an exit
        message arrives or a timeout occurs.
        """


def setup_controller(zmq_context=None):
    if not zmq_context:
        import zmq
        zmq_context = zmq.Context()
    controller = Controller(zmq_context=zmq_context)
    return controller



import pytest

import subprocess
import sys
import tempfile

from pelita.network import ZMQClientError
from pelita.scripts.pelita_tournament import firstNN
from pelita.tournament import call_pelita, check_team

_mswindows = (sys.platform == "win32")


class TestLibpelitaUtils:
    def test_firstNN(self):
        assert firstNN(None, False, True) == False
        assert firstNN(True, False, True) == True
        assert firstNN(None, None, True) == True
        assert firstNN(None, 2, True) == 2
        assert firstNN(None, None, None) == None
        assert firstNN() == None

def test_call_pelita():
    rounds = 200
    viewer = 'ascii'
    size = 'small'

    teams = ["pelita/player/StoppingPlayer", "pelita/player/StoppingPlayer"]
    (state, stdout, stderr) = call_pelita(teams, rounds=rounds, viewer='null', size=size, seed=None)
    assert state['gameover'] is True
    assert state['whowins'] == 2
    # Quick assert that there is text in stdout
    assert len(stdout.split('\n')) > 0
    assert "Finished" in stdout

    teams = ["pelita/player/SmartEatingPlayer", "pelita/player/StoppingPlayer"]
    (state, stdout, stderr) = call_pelita(teams, rounds=rounds, viewer=viewer, size=size, seed=None)
    assert state['gameover'] is True
    assert state['whowins'] == 0

    teams = ["pelita/player/StoppingPlayer", "pelita/player/SmartEatingPlayer"]
    (state, stdout, stderr) = call_pelita(teams, rounds=rounds, viewer=viewer, size=size, seed=None)
    assert state['gameover'] is True
    assert state['whowins'] == 1


@pytest.mark.parametrize('seed,success', [
        (None, True),
        (0, True),
        (1, True),
        ("1", True),
        ("193294814091830945831093", True),
        ("1.0", False),
        ("daflj", False),
])
def test_bad_seeds(seed, success):
    rounds = 2
    viewer = 'null'
    size = 'small'

    teams = ["pelita/player/StoppingPlayer", "pelita/player/StoppingPlayer"]
    if success:
        (state, stdout, stderr) = call_pelita(teams, rounds=rounds, viewer='null', size=size, seed=seed)
        assert state['gameover'] is True
    else:
        with pytest.raises(ValueError):
                call_pelita(teams, rounds=rounds, viewer='null', size=size, seed=seed)


def test_check_team_external():
    assert check_team("pelita/player/StoppingPlayer") == "Stopping Players"

def test_check_team_external_fails():
    with pytest.raises(ZMQClientError):
        check_team("Unknown Module")

def test_check_team_internal():
    def move(b, s):
        return b.position, s
    assert check_team(move) == "local-team (move)"


@pytest.mark.skipif(_mswindows, reason="NamedTemporaryFiles cannot be used in another process")
def test_write_replay_is_idempotent():
    # TODO: The replay functionality could be added to call_pelita
    # so we don’t have to run the subprocess ourselves
    with tempfile.NamedTemporaryFile() as f:
        with tempfile.NamedTemporaryFile() as g:
            # run a quick game and save the game states to f

            cmd = [sys.executable, '-m', 'pelita.scripts.pelita_main',
                    '--write-replay', f.name,
                    '--size', 'small',
                    '--null']

            subprocess.run(cmd, check=True)

            f.seek(0)
            first_run = f.read()
            # check that we received something
            assert len(first_run) > 0

            # run a replay of f and store in g
            cmd = [sys.executable, '-m', 'pelita.scripts.pelita_main',
                    '--write-replay', g.name,
                    '--replay', f.name,
                    '--null']

            subprocess.run(cmd, check=True)

            g.seek(0)
            second_run = g.read()
            # check that f and g have the same content
            assert first_run == second_run




import itertools
import math
import queue
from collections import defaultdict, namedtuple
from io import StringIO


def sort_ranks(teams, bonusmatch=False):
    """ Re-orders a ranked list of teams such that
    the best and worst team are next to each other,
    the second best and second best are next to each other, etc.

    If bonusmatch is True, then the worst team will be left
    out until the final match

    Parameters
    ----------
    teams : ranked list of teams

    Raises
    ------
    TypeError
        if new_data is not a list
    ValueError
        if new_data has inappropriate length

    """
    if len(teams) < 2:
        return teams

    if bonusmatch:
        # pop last item from teams list
        bonus_team = [teams.pop()]
    else:
        bonus_team = []

    # if the number of teams is not even, we need to leave out another team
    l = len(teams)
    if l % 2 != 0:
        # pop last item from teams list
        remainder_team = [teams.pop()]
    else:
        remainder_team = []

    # top half
    good_teams = teams[:l//2]
    # bottom half
    bad_teams = reversed(teams[l//2:2*(l//2)])

    pairs = [team for pair in zip(good_teams, bad_teams)
                  for team in pair]
    return pairs + remainder_team + bonus_team

def identity(x):
    return x

class MatrixElem:
    def size(self, trafo=identity):
        return len(self.to_s(trafo=trafo))

    def box(self, team, *, prefix=None, postfix=None, size=None, padLeft="", padRight="", fillElem="─", highlighted=False):
        if prefix is None:
            prefix = ""
        if postfix is None:
            postfix = ""

        if size is None:
            size = 0
        else:
            size = size - len(prefix) - len(postfix)

        BOLD = '\033[1m'
        END = '\033[0m'

        padded = "{padLeft}{team}{padRight}".format(team=team, padLeft=padLeft, padRight=padRight)
        return "{prefix}{BOLD}{team:{fillElem}<{size}}{END}{postfix}".format(team=padded, prefix=prefix, postfix=postfix,
                                                                             size=size, fillElem=fillElem,
                                                                             BOLD=BOLD if highlighted else "",
                                                                             END=END if highlighted else "")

class Team(namedtuple("Team", ["name"]), MatrixElem):
    def to_s(self, size=None, trafo=identity, highlighted=False):
        return self.box(trafo(self.name), size=size, prefix="", padLeft=" ", padRight=" ", highlighted=highlighted)

class Bye(namedtuple("Bye", ["team"]), MatrixElem):
    def to_s(self, size=None, trafo=identity, highlighted=False):
        prefix = "──"
        # return show_team("…", prefix=prefix, padLeft=" ", padRight=" ", size=size)
        return self.box("", size=size)

class Match(namedtuple("Match", ["t1", "t2"]), MatrixElem):
    def __init__(self, *args, **kwargs):
        self.winner = None

    def __repr__(self):
        return "Match(t1={}, t2={}, winner={})".format(self.t1, self.t2, self.winner)

    def to_s(self, size=None, trafo=identity, highlighted=False):
        prefix = "├─"
        name = trafo(self.winner) if (self.winner is not None) else "???"
        return self.box(name, prefix=prefix, padLeft=" ", padRight=" ", size=size, highlighted=highlighted)

class FinalMatch(namedtuple("FinalMatch", ["t1", "t2"]), MatrixElem):
    def __init__(self, *args, **kwargs):
        self.winner = None
    def to_s(self, size=None, trafo=identity, highlighted=False):
        prefix = "├──┨"
        postfix = "┃"
        fillElem = " "
        name = trafo(self.winner) if (self.winner is not None) else "???"
        return self.box(name, prefix=prefix, postfix=postfix, padLeft=" ", padRight=" ", fillElem=fillElem, size=size, highlighted=highlighted)

class Element(namedtuple("Element", ["char"]), MatrixElem):
    def to_s(self, size=None, trafo=identity, highlighted=False):
        return self.box(self.char, size=size, fillElem=" ", highlighted=highlighted)

class Empty(namedtuple("Empty", []), MatrixElem):
    def to_s(self, size=None, trafo=identity, highlighted=False):
        return self.box(" ", size=size, fillElem=" ")

class BorderTop(namedtuple("BorderTop", ["team", "tight"]), MatrixElem):
    def to_s(self, size=None, trafo=identity, highlighted=False):
        prefix = "│  " if not self.tight else "┐  "
        padRight = ""
        padLeft = "┏"
        postfix = "┓"
        fillElem = "━"
        return self.box("", prefix=prefix, postfix=postfix, padLeft=padLeft, padRight=padRight, fillElem=fillElem, size=size)

class BorderBottom(namedtuple("BorderBottom", ["team", "tight"]), MatrixElem):
    def to_s(self, size=None, trafo=identity, highlighted=False):
        prefix = "│  " if not self.tight else "┘  "
        padRight = ""
        padLeft = "┗"
        postfix = "┛"
        fillElem = "━"
        return self.box("", prefix=prefix, postfix=postfix, padLeft=padLeft, padRight=padRight, fillElem=fillElem, size=size)

def knockout_matrix(tree):
    """
    For now teams is a list (cols) of list (rows) of teams
    """
    teams = tree_enumerate(tree)

    initial_teams = teams[0]
    N = len(initial_teams)
    height = N * 2 - 1
    width = len(teams)

    matrix = [[Empty() for _w in range(width)] for _h in range(height)]
    # fill left column with initial teams
    for idx, t in enumerate(initial_teams):
        matrix[idx * 2][0] = t

    last_match = None

    for g_idx, generation in enumerate(teams):
        # print("G:", generation)
        if g_idx == 0:
            continue
        col = g_idx
        left_col = g_idx - 1
        for m_idx, match in enumerate(generation):
        #    print("M:", match)
            if isinstance(match, Match):
                # find row idx of the match partners
                for row_idx, row in enumerate(matrix):
                    if row[left_col] == match.t1:
                        start_row = row_idx
                    if row[left_col] == match.t2:
                        end_row = row_idx
                middle_row = math.floor(start_row + (end_row - start_row) / 2)

                # draw next match
                for row in range(start_row, end_row):
                    matrix[row][col] = Element('│')
                matrix[start_row][col] = Element('┐')
                matrix[end_row][col] = Element('┘')
                matrix[middle_row][col] = match
                last_match = (middle_row, col)

            if isinstance(match, Bye):
                for row_idx, row in enumerate(matrix):
                    if row[left_col] == match.team:
                        break
                matrix[row_idx][col] = match

    return matrix, last_match

def print_knockout(tree, name_trafo=identity, highlight=None):
    if highlight is None:
        highlight = []

    matrix, final_match = knockout_matrix(tree)

    winner_row = final_match[0]
    winning_team = matrix[final_match[0]][final_match[1]].winner
    winner = matrix[final_match[0]][final_match[1]] = FinalMatch(* matrix[final_match[0]][final_match[1]])
    winner.winner = winning_team

    def is_tight(elem):
        return not isinstance(elem, Empty) and not isinstance(elem, Element)

    matrix[winner_row - 1][-1] = BorderTop(winner, is_tight(matrix[winner_row - 1][-2]))
    matrix[winner_row + 1][-1] = BorderBottom(winner, is_tight(matrix[winner_row + 1][-2]))

    # estimate the width that a column needs
    colwidths = [0] * len(matrix[0])
    for row in matrix:
        for col_idx, elem in enumerate(row):
            current_width = elem.size(trafo=name_trafo)
            old_width = colwidths[col_idx]
            colwidths[col_idx] = max(current_width, old_width)

    with StringIO() as output:
        for row in range(len(matrix)):
            for col in range(len(matrix[0])):
                try:
                    elem = matrix[row][col]

                    str = elem.to_s(colwidths[col], trafo=name_trafo, highlighted=elem in highlight)
                    print(str, end="", file=output)
                except AttributeError:
                    print("Here:", end="")
                    print(row, col, matrix[row][col])
                    raise
            print(file=output)
        return output.getvalue()


def makepairs(matches):
    if len(matches) == 0:
        raise ValueError("Cannot prepare matches (no teams given).")
    while not len(matches) == 1:
        m = []
        pairs = itertools.zip_longest(matches[::2], matches[1::2])
        for p1, p2 in pairs:
            if p2 is not None:
                m.append(Match(p1, p2)) #  winner=None))
            else:
                m.append(Bye(p1))
        matches = m
    return matches[0]

def prepare_matches(teams, bonusmatch=False):
    """ Takes a ranked list of teams, matches them according to sort_ranks
    and returns the Match tree.
    """

    if not teams:
        raise ValueError("No teams given to sort.")

    teams_sorted = sort_ranks(teams, bonusmatch=bonusmatch)

    # If there is a bonus match, we must ensure that it will be played
    # at the very last
    if bonusmatch:
        bonus_team = teams_sorted.pop()
        if not teams_sorted:
            return Team(bonus_team)

    # pair up the games and return the tree starting from the winning team
    match_tree = makepairs([Team(t) for t in teams_sorted])

    if bonusmatch:
        # now add enough Byes to the bonus_team
        # so that we still have a balanced tree
        # when we add the bonus_team as a final match
        team = Team(bonus_team)
        for _depth in range(tree_depth(match_tree) - 1):
            team = Bye(team)

        match_tree = Match(match_tree, team)

    # ensure we have a balanced tree
    assert is_balanced(match_tree)

    return match_tree

def is_balanced(tree):
    if isinstance(tree, Match):
        return is_balanced(tree.t1) and is_balanced(tree.t2) and tree_depth(tree.t1) == tree_depth(tree.t2)
    if isinstance(tree, Bye):
        return True
    if isinstance(tree, Team):
        return True

def tree_depth(tree):
    if isinstance(tree, Match):
        return 1 + max(tree_depth(tree.t1), tree_depth(tree.t2))
    if isinstance(tree, Bye):
        return 1 + tree_depth(tree.team)
    if isinstance(tree, Team):
        return 1

def tree_enumerate(tree):
    enumerated = defaultdict(list)

    nodes = queue.Queue()
    nodes.put((tree, 0))
    while not nodes.empty():
        node, generation = nodes.get()
        if isinstance(node, Match):
            nodes.put((node.t1, generation + 1))
            nodes.put((node.t2, generation + 1))
        if isinstance(node, Bye):
            nodes.put((node.team, generation + 1))
        if isinstance(node, Team):
            pass
        enumerated[generation].append(node)

    generations = []
    for idx in sorted(enumerated.keys()):
        generations.append(enumerated[idx])
    generations.reverse()
    return generations



#!/usr/bin/env python3

# Copyright (c) 2013, Bastian Venthur <venthur@debian.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Continuous Integration Engine.

Currently this module is only usable as a script, later it shall be
extended to be used as a library for a web service providing a
continuous integration service.

In its current form it is still very usable to test a couple of agents
automatically against each other and compare the results. For best
results modify the ``agents`` section in the ``ci.cfg`` configuration
file and run this file. Leave it running for a while until the positions
stabilized.

"""


import configparser
import itertools
import json
import logging
import random
import sqlite3
import subprocess
import sys

import click
from rich.console import Console
from rich.table import Table

from pelita.network import ZMQClientError
from pelita.tournament import check_team, call_pelita
from pelita.scripts.script_utils import start_logging

_logger = logging.getLogger(__name__)

# the path of the configuration file
CFG_FILE = './ci.cfg'

def hash_team(team_spec):
    external_call = [sys.executable,
                    '-m',
                    'pelita.scripts.pelita_player',
                    'hash-team',
                    team_spec]
    _logger.debug("Executing: %r", external_call)
    res = subprocess.run(external_call, capture_output=True, text=True)
    return res.stdout.strip()

class CI_Engine:
    """Continuous Integration Engine."""

    def __init__(self, cfgfile):
        self.players = []
        config = configparser.ConfigParser()
        config.read_file(cfgfile)
        for name, path in  config.items('agents'):
            self.players.append({'name': name, 'path': path})

        self.rounds = config['general'].getint('rounds', None)
        self.size = config['general'].get('size', None)
        self.viewer = config['general'].get('viewer', 'null')
        self.seed = config['general'].get('seed', None)

        self.db_file = config.get('general', 'db_file')
        self.dbwrapper = DB_Wrapper(self.db_file)

    def load_players(self):
        hash_cache = {}

        # remove players from db which are not in the config anymore
        for pname in self.dbwrapper.get_players():
            if pname not in [p['name'] for p in self.players]:
                _logger.debug('Removing %s from database, because it is not among the current players.' % (pname))
                self.dbwrapper.remove_player(pname)

        # add new players into db
        for player in self.players:
            pname, path = player['name'], player['path']
            if pname not in self.dbwrapper.get_players():
                _logger.debug('Adding %s to database.' % pname)
                hash_cache[path] = hash_team(path)
                self.dbwrapper.add_player(pname, hash_cache[path])

        # reset players where the directory hash changed
        for player in self.players:
            path = player['path']
            pname = player['name']
            new_hash = hash_cache.get(path, hash_team(path))
            if new_hash != self.dbwrapper.get_player_hash(pname):
                _logger.debug('Resetting %s because its module hash changed.' % pname)
                self.dbwrapper.remove_player(pname)
                self.dbwrapper.add_player(pname, new_hash)

        for player in self.players:
            path = player['path']
            pname = player['name']
            try:
                _logger.debug('Querying team name for %s.' % pname)
                team_name = check_team(player['path'])
                self.dbwrapper.add_team_name(pname, team_name)
            except ZMQClientError as e:
                e_type, e_msg = e.args
                _logger.debug(f'Could not import {player} at path {path} ({e_type}): {e_msg}')
                player['error'] = e.args

    def run_game(self, p1, p2):
        """Run a single game.

        This method runs a single game ``p1`` vs ``p2`` and internally
        stores the result.

        Parameters
        ----------
        p1, p2 : int
            the indices of the players

        """
        team_specs = [self.players[i]['path'] for i in (p1, p2)]
        print(f"Playing {self.players[p1]['name']} against {self.players[p2]['name']}.")

        final_state, stdout, stderr = call_pelita(team_specs,
                                                            rounds=self.rounds,
                                                            size=self.size,
                                                            viewer=self.viewer,
                                                            seed=self.seed)

        if final_state['whowins'] == 2:
            result = -1
        else:
            result = final_state['whowins']

        del final_state['walls']
        del final_state['food']

        _logger.info('Final state: %r', final_state)
        _logger.debug('Stdout: %r', stdout)
        if stderr:
            _logger.warning('Stderr: %r', stderr)
        p1_name, p2_name = self.players[p1]['name'], self.players[p2]['name']
        self.dbwrapper.add_gameresult(p1_name, p2_name, result, final_state, stdout, stderr)


    def start(self, n):
        """Start the Engine.

        This method will start and infinite loop, testing each agent
        randomly against another one. The result is printed after each
        game.

        Currently the only way to stop the engine is via CTRL-C.

        Examples
        --------
        >>> ci = CI_Engine()
        >>> ci.start()

        """
        loop = itertools.repeat(None) if n == 0 else itertools.repeat(None, n)

        for _ in  loop:
            # choose the player with the least number of played game,
            # mix him with another random player
            # mis the sides and let them play
            broken_players = {idx for idx, player in enumerate(self.players) if player.get('error')}
            game_count = [(self.dbwrapper.get_game_count(p['name']), idx) for idx, p in enumerate(self.players)]
            players_sorted = [idx for count, idx in sorted(game_count) if not idx in broken_players]
            a, rest = players_sorted[0], players_sorted[1:]
            b = random.choice(rest)
            players = [a, b]
            random.shuffle(players)

            self.run_game(players[0], players[1])
            self.pretty_print_results(highlight=[self.players[players[0]]['name'], self.players[players[1]]['name']])
            print('------------------------------')


    def get_results(self, idx, idx2=None):
        """Get the results so far.

        This method goes through the internal list of of all game
        results and calculates the result for the player with index
        ``idx`` against everyone else.

        If the optional argument ``idx2`` is given only the results of
        the players ``idx`` vs ``idx2`` are returned.

        Parameters
        ----------
        idx : int
            the index of the player
        idx2 : int, optional
            the index of the second player if this parameter is not
            given the result of player against all other players is
            returned otherwise the results of the games of the players
            with the indices ``idx`` and ``idx2`` are returned


        Returns
        -------
        win, loss, draw : int
            the number of wins, losses and draws for this player or
            combination of players


        Examples
        --------

        >>> # get the results of player with index 1 against all other
        >>> # players
        >>> ci.get_results(1)
        (5, 2, 0)
        >>> # get the results of all games with the players of index 1
        >>> # and 5
        >>> ci.get_results(1, 5)
        (2, 0, 0)

        """
        win, loss, draw = 0, 0, 0
        p1_name = self.players[idx]['name']
        p2_name = None if idx2 == None else self.players[idx2]['name']
        relevant_results = self.dbwrapper.get_results(p1_name, p2_name)
        for p1, p2, r in relevant_results:
            if (idx2 is None and p1_name == p1) or (idx2 is not None and p1_name == p1 and p2_name == p2):
                if r == 0: win += 1
                elif r == 1: loss += 1
                elif r == -1: draw += 1
            if (idx2 is None and p1_name == p2) or (idx2 is not None and p1_name == p2 and p2_name == p1):
                if r == 1: win += 1
                elif r == 0: loss += 1
                elif r == -1: draw += 1
        return win, loss, draw

    def get_errorcount(self, idx):
        """Gets the error count for team idx

        Parameters
        ----------
        idx : int
            the index of the player

        Returns
        -------
        error_count, fatalerror_count : int
            the number of errors for this player

        """
        p_name = self.players[idx]['name']
        error_count, fatalerror_count = self.dbwrapper.get_errorcount(p_name)
        return error_count, fatalerror_count

    def get_team_name(self, idx):
        """Get last registered team name.

        team_name : string
        """

        p_name = self.players[idx]['name']
        return self.dbwrapper.get_team_name(p_name)

    def gen_elo(self):
        k = 32

        def elo_change(a, b, outcome):
            expected = 1 / ( 10**((b - a) / 400) + 1 )
            return k * (outcome - expected)

        from collections import defaultdict
        elo = defaultdict(lambda: 1500)

        g = self.dbwrapper.cursor.execute("""
        SELECT player1, player2, result
        FROM games
        """).fetchall()
        for p1, p2, result in g:
            if result == 0:
                change = elo_change(elo[p1], elo[p2], 1)
            if result == 1:
                change = elo_change(elo[p1], elo[p2], 0)
            if result == -1:
                change = elo_change(elo[p1], elo[p2], 0.5)
            elo[p1] += change
            elo[p2] -= change

        return elo

    def pretty_print_results(self, highlight=None):
        """Pretty print the current results.

        """
        if highlight is None:
            highlight = []

        console = Console()
        # Some guesswork in here
        MAX_COLUMNS = (console.width - 40) // 12
        if MAX_COLUMNS < 4:
            # Let’s be honest: You should enlarge your terminal window even before that
            MAX_COLUMNS = 4

        res = self.dbwrapper.get_wins_losses()
        rows = { k: list(v) for k, v in itertools.groupby(res, key=lambda x:x[0]) }

        good_players = [p for p in self.players if not p.get('error')]
        bad_players = [p for p in self.players if p.get('error')]

        num_rows_per_player = (len(good_players) // MAX_COLUMNS) + 1
        row_style = [*([""] * num_rows_per_player), *(["dim"] * num_rows_per_player)]

        table = Table(row_styles=row_style, title="Cross results")
        table.add_column("")
        table.add_column("Name")
        table.add_column("Score", justify="right")
        table.add_column("W/D/L")

        column_players = [[] for _idx in range(min(MAX_COLUMNS, len(good_players)))]
        # if we have more good_players than allowed columns, we must wrap around
        for idx, _p in enumerate(good_players):
            column_players[idx % MAX_COLUMNS].append(idx)

        for midx in column_players:
            table.add_column('\n'.join(map(str, midx)))


        def batched(iterable, n):
            # Backport from Python 3.12
            # batched('ABCDEFG', 3) → ABC DEF G
            if n < 1:
                raise ValueError('n must be at least one')
            iterator = iter(iterable)
            while batch := tuple(itertools.islice(iterator, n)):
                yield batch

        result = []
        for idx, p in enumerate(good_players):
            win, loss, draw = self.get_results(idx)
            error_count, fatalerror_count = self.get_errorcount(idx)
            try:
                team_name = self.get_team_name(idx)
            except ValueError:
                team_name = None
            score = 0 if (win+loss+draw) == 0 else (win-loss) / (win+loss+draw)
            result.append([score, win, draw, loss, p['name'], team_name, error_count, fatalerror_count])
            wdl = f"{win:3d},{draw:3d},{loss:3d}"

            try:
                row = rows[p['name']]
            except KeyError:
                continue
            vals = { k: (w,l,d) for _p1, k, w, l, d in row }

            cross_results = []
            for idx2, p2 in enumerate(good_players):
                win, loss, draw = vals.get(p2['name'], (0, 0, 0))
                if idx == idx2:
                    cross_results.append("  - - - ")
                else:
                    cross_results.append(f"{win:2d},{draw:2d},{loss:2d}")

            for c, r in enumerate(batched(cross_results, MAX_COLUMNS)):
                if c == 0:
                    table.add_row(f"{idx}", p['name'], f"{score:.2f}", wdl, *r)
                else:
                    table.add_row("", "", "", "", *r)

        console.print(table)

        table = Table(title="Bot ranking")

        table.add_column("Name")
        table.add_column("# Matches")
        table.add_column("# Wins")
        table.add_column("# Draws")
        table.add_column("# Losses")
        table.add_column("Score")
        table.add_column("ELO")
        table.add_column("Error count")
        table.add_column("# Fatal Errors")

        elo = self.gen_elo()

        result.sort(reverse=True)
        for [score, win, draw, loss, name, team_name, error_count, fatalerror_count] in result:
            style = "bold" if name in highlight else None
            display_name = f"{name} ({team_name})" if team_name else f"{name}"
            table.add_row(
                display_name,
                f"{win+draw+loss}",
                f"{win}",
                f"{draw}",
                f"{loss}",
                f"{score:6.3f}",
                f"{elo[name]: >4.0f}",
                f"{error_count}",
                f"{fatalerror_count}",
                style=style,
            )

        console.print(table)

        for p in bad_players:
            print("% 30s ***%30s***" % (p['name'], p['error']))


class DB_Wrapper:
    """Wrapper around the games database."""

    def __init__(self, dbfile):
        """Initialize the connection to the db ``dbfile``.

        Create table if file does not exist.

        Parameters
        ----------
        dbfile : str
            path to sqlite3 database

        """
        self.db_file = dbfile
        self.connection = sqlite3.connect(self.db_file)
        self.cursor = self.connection.cursor()
        self.cursor.execute("PRAGMA foreign_keys = ON;")
        self.create_tables()

    def create_tables(self):
        """Create tables.

        This is a no-op if the tables already exist.

        """
        self.cursor.execute("""
        CREATE TABLE IF NOT EXISTS players
        (name text PRIMARY KEY, hash text)
        """)
        self.cursor.execute("""
        CREATE TABLE IF NOT EXISTS team_names
        (name text PRIMARY KEY, team_name text,
        FOREIGN KEY(name) REFERENCES players(name) ON DELETE CASCADE)
        """)
        self.cursor.execute("""
        CREATE TABLE IF NOT EXISTS games
        (player1 text, player2 text, result int, final_state text, stdout text, stderr text,
        FOREIGN KEY(player1) REFERENCES players(name) ON DELETE CASCADE,
        FOREIGN KEY(player2) REFERENCES players(name) ON DELETE CASCADE)
        """)
        self.connection.commit()

    def get_players(self):
        """Get players from the database.

        Returns
        -------
        players : list of strings
            the player names from the database.

        """
        players = self.cursor.execute("""SELECT name FROM players""").fetchall()
        players = [row[0] for row in players]
        return players

    def get_player_hash(self, name):
        """Get the hash stored in the database for the player.

        Raises
        ------
        ValueError : if the player does not exist in the database

        """
        h = self.cursor.execute("""
        SELECT hash
        FROM players
        WHERE name = ?
        """, (name,)).fetchone()
        if h is None:
            raise ValueError('Player %s does not exist in database.' % name)
        return h[0]

    def add_player(self, name, h):
        """Add player to database

        Parameters
        ----------
        name : str
        h : str
            hash of the player's directory

        Raises
        ------
        ValueError : if player already exists in database

        """
        try:
            self.cursor.execute("""
            INSERT INTO players
            VALUES (?, ?)
            """, [name, h])
            self.connection.commit()
        except sqlite3.IntegrityError:
            raise ValueError('Player %s already exists in database' % name)

    def add_team_name(self, name, team_name):
        """Adds or updates team name to database

        Parameters
        ----------
        name : str
        team_name : str

        """
        try:
            self.cursor.execute("""
            INSERT OR REPLACE INTO team_names
            VALUES (?, ?)
            """, [name, team_name])
            self.connection.commit()
        except sqlite3.IntegrityError:
            raise ValueError('Cannot add team name for %s' % name)

    def remove_player(self, pname):
        """Remove a player from the database.

        Removes all games where the player ``pname`` participated.

        Parameters
        ----------
        pname : str
            the player name of the player to be removed

        """
        self.cursor.execute("""DELETE FROM games
        WHERE player1 = ? or player2 = ?""", (pname, pname))
        self.cursor.execute("""DELETE FROM players
        WHERE name = ?""", (pname,))
        self.connection.commit()

    def add_gameresult(self, p1_name, p2_name, result, final_state, std_out, std_err):
        """Add a new game result to the database.

        Parameters
        ----------
        p1_name, p2_name : str
            the names of the players
        result : 0, 1 or -1
            0 if player 1 won
            1 of player 2 won
            -1 if draw
        std_out, std_err : str
            STDOUT and STDERR of the game

        """
        self.cursor.execute("""
        INSERT INTO games
        VALUES (?, ?, ?, ?, ?, ?)
        """, [p1_name, p2_name, result, json.dumps(final_state), std_out, std_err])
        self.connection.commit()

    def get_results(self, p1_name, p2_name=None):
        """Get all games involving player1 (AND player2 if specified).

        Parameters
        ----------
        p1_name : str
            the  name of player 1
        p2_name : str, optional
            the name of player 2, if not specified ``get_results`` will
            return all games involving player 1 otherwise it will return
            all games of player1 AND player2

        Returns
        -------
        relevant_results : list of gameresults

        """
        if p2_name is None:
            self.cursor.execute("""
            SELECT player1, player2, result FROM games
            WHERE player1 = ? or player2 = ?""", (p1_name, p1_name))
            relevant_results = self.cursor.fetchall()
        else:
            self.cursor.execute("""
            SELECT player1, player2, result FROM games
            WHERE (player1 = :p1 and player2 = :p2) or (player1 = :p2 and player2 = :p1)""",
            dict(p1=p1_name, p2=p2_name))
            relevant_results = self.cursor.fetchall()
        return relevant_results


    def get_team_name(self, p_name):
        """Gets the last registered team name of p_name.

        Parameters
        ----------
        p_name : str
            the name of the player

        Returns
        -------
        team_name : str

        """
        self.cursor.execute("""
        SELECT team_name FROM team_names
        WHERE name = ?""", (p_name,))
        res = self.cursor.fetchone()
        if res is None:
            raise ValueError('Player %s does not exist in database.' % p_name)
        return res[0]

    def get_game_count(self, p1_name, p2_name=None):
        """Get number of games involving player1 (AND player2 if specified).

        Parameters
        ----------
        p1_name : str
            the  name of player 1
        p2_name : str, optional
            the name of player 2, if not specified ``get_results`` will
            return all games involving player 1 otherwise it will return
            all games of player1 AND player2

        Returns
        -------
        relevant_results : list of gameresults

        """
        if p2_name is None:
            self.cursor.execute("""
            SELECT count(*) FROM games
            WHERE player1 = ? or player2 = ?""", (p1_name, p1_name))
            count, = self.cursor.fetchone()
        else:
            self.cursor.execute("""
            SELECT count(*) FROM games
            WHERE (player1 = :p1 and player2 = :p2) or (player1 = :p2 and player2 = :p1)""",
            dict(p1=p1_name, p2=p2_name))
            count, = self.cursor.fetchone()
        return count

    def get_errorcount(self, p1_name):
        """Get errorcount of player1

        Parameters
        ----------
        p1_name : str
            the  name of player 1

        Returns
        -------
        error_count, fatalerror_count : errorcount

        """
        self.cursor.execute("""
        SELECT sum(c) FROM
            (
                SELECT sum(json_extract(final_state, '$.num_errors[0]')) AS c
                FROM games
                WHERE player1 = :p1

                UNION ALL

                SELECT sum(json_extract(final_state, '$.num_errors[1]')) AS c
                FROM games
                WHERE player2 = :p1
            )
        """,
        dict(p1=p1_name))
        error_count, = self.cursor.fetchone()

        self.cursor.execute("""
        SELECT sum(c) FROM
            (
                SELECT count(*) AS c
                FROM games
                WHERE player1 = :p1 AND
                      json_extract(final_state, '$.fatal_errors[0]') != '[]'

                UNION ALL

                SELECT count(*) AS c
                FROM games
                WHERE player2 = :p1 AND
                      json_extract(final_state, '$.fatal_errors[1]') != '[]'
            )
        """,
        dict(p1=p1_name))
        fatal_errorcount, = self.cursor.fetchone()

        return error_count, fatal_errorcount

    def get_wins_losses(self, team=None):
        """ Get all wins and losses combined in a table of
        team | opponent | wins | losses | draws
        """

        if team is not None:
            where_clause = "WHERE team = ?"
        else:
            where_clause = ""

        query = f"""

        SELECT
            team, opponent, SUM(wins) AS wins, SUM(losses) AS losses, SUM(draws) AS draws
        FROM (
            -- Count wins for player1
            SELECT
                player1 AS team, player2 AS opponent, COUNT(*) AS wins, 0 AS losses, 0 AS draws
            FROM games
            WHERE result = 0
            GROUP BY player1, player2

            UNION ALL

            -- Count wins for player2
            SELECT
                player2 AS team, player1 AS opponent, 0 AS wins, COUNT(*) AS losses, 0 AS draws
            FROM games
            WHERE result = 0
            GROUP BY player2, player1

            UNION ALL

            -- Count losses for player1
            SELECT
                player1 AS team, player2 AS opponent, 0 AS wins, COUNT(*) AS losses, 0 AS draws
            FROM games
            WHERE result = 1
            GROUP BY player1, player2

            UNION ALL

            -- Count losses for player2
            SELECT
                player2 AS team, player1 AS opponent, COUNT(*) AS wins, 0 AS losses, 0 AS draws
            FROM games
            WHERE result = 1
            GROUP BY player2, player1

            UNION ALL

            -- Count draws for both teams
            SELECT
                player1 AS team, player2 AS opponent, 0 AS wins, 0 AS losses, COUNT(*) AS draws
            FROM games
            WHERE result = -1
            GROUP BY player1, player2

            UNION ALL

            SELECT
                player2 AS team, player1 AS opponent, 0 AS wins, 0 AS losses, COUNT(*) AS draws
            FROM games
            WHERE result = -1
            GROUP BY player2, player1
        ) AS results
        {where_clause}
        GROUP BY
            team, opponent
        ORDER BY
            team, opponent
        ;
        """
        if team is not None:
            return self.cursor.execute(query, [team]).fetchall()
        else:
            return self.cursor.execute(query).fetchall()


@click.command()
@click.option('--log',
              is_flag=False, flag_value="-", default=None, metavar='LOGFILE',
              help="print debugging log information to LOGFILE (default 'stderr')")
@click.option('--config',
              default=CFG_FILE,
              type=click.File('r'),
              help='Configuration file')
@click.option('-n', help='run N times', type=int, default=0)
@click.option('--print', is_flag=True, default=False,
              help='Print scores and exit.')
@click.option('--nohash', is_flag=True, default=False,
              help='Do not hash the players')
def main(log, config, n, print, nohash):
    if log is not None:
        start_logging(log, __name__)
        start_logging(log, 'pelita')

    ci_engine = CI_Engine(config)
    if print:
        ci_engine.pretty_print_results()
    else:
        if not nohash:
            ci_engine.load_players()
        ci_engine.start(n)

if __name__ == '__main__':
    main()
